/* $Id: rxhell18.cpp,v 1.3 1995/09/23 21:19:41 artc Exp $ */
/* 
      (C) Copyright 1993-1995 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

//
// rxhell18.cpp - Rx app to create and list block definitions
//
// Jeremy Tammik, Autodesk ESC, 94-08-15
//
// RxHello18 creates and lists block definitions
// in the AutoCAD database.
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h> // strupr()
#include <aced.h>
#include <dbsymtb.h>
#include <dbents.h>
#include <adslib.h>
#include "rxhell18.h"
#include "adeskabb.h"

#define streq(s,t) (strcmp((s),(t))==0)

#ifdef sgi
/************************************************************************/
/*									*/
/* convert string to uppercase characters 				*/
/*									*/
/************************************************************************/
char *strupr(char *str)
{
	char *s = str;

	if (! *s)
	{
		return(str);
	}

	do
	{
		if ('a' <= *s && *s <= 'z')
		{
			*s += ('A'-'a');
		}
	}
	while(*s++);

	return(str);
}
#endif /* sgi */

// hello18 - AutoCAD command to manipulate block definitions:
//
// Creates, iterates and modifies block definitions
// in the AutoCAD database.
//
void
hello18(void)
{
    ads_printf("This is hello18() in RxHello18 ...\n");

    char kw[20];
    ads_initget(0, "Create List");
    int rc = ads_getkword("Create/<List>: ", kw);

    if (rc == RTNONE) {
        strcpy(kw, "List");
        rc = RTNORM;
    }

    if (rc != RTNORM) {
        ads_printf("Nothing selected.\n");
        return;
    }

    if (streq(kw, "List")) {
        listBlock();
    } else if (streq(kw, "Create")) {
        createBlock();
    }
    return;
}

// createBlock - create a block definition using selected entites:
//
// The selected entities are cloned, inserted into the block
// definition, and deleted from the database, just like the 
// AutoCAD BLOCK command does.
//
Boolean
createBlock()
{
    // get name of block:
    //
    char bName[80];
    int rc = ads_getstring(kFalse, "Enter block name: ", bName);

    if (rc!=RTNORM || *bName == '\0') {
        ads_printf("Invalid block name '%s'.\n", bName);
        return kFalse;
    }

    // get entities for block:
    //
    ads_name ss;
    rc = ads_ssget(NULL, NULL, NULL, NULL, ss);

    if (rc!=RTNORM) {
        ads_printf("Nothing selected.\n");
        return kFalse;
    }

    long len;
    rc = ads_sslength(ss, &len);

    if (rc!=RTNORM) {
        ads_printf("Invalid selection set.\n");
        ads_ssfree(ss);
        return kFalse;
    }

    // get insertion point of block:
    //
    AcGePoint3d p;
    rc = ads_getpoint(NULL, "Block insertion point <0,0,0>: ", 
    (double*)(&p));

    if (rc==RTNONE) {
        p[X] = p[Y] = p[Z] = 0.0;
        rc = RTNORM;
    }

    if (rc!=RTNORM) {
        ads_printf("Invalid insertion point.\n");
        return kFalse;
    }

    // create and name a new block table record:
    //
    AcDbBlockTableRecord *pBlockTableRec = new AcDbBlockTableRecord();

    // set block table record name to bName and block origin to p:
    //
    // ...

    // get block symbol table:
    //
    AcDbBlockTable *pBlockTable = NULL;

    // retrieve block table from database, open for writing:
    //
    // ...

    AcDbObjectId blockTableRecordId;

    // add block table record to block table:
    //
    // ...

    // close block table now that the new block record is added:
    //
    // ...

    // add entities to the block record:
    //
    ads_name en;
    AcDbObjectId objId;
    AcDbEntity *pObj, *pClone;
    Boolean myrc = kTrue;

    for (long i=0; i<len; ++i) {
        ads_ssname(ss, i, en);
        if (acdbGetObjectId(objId, en) != Acad::eOk) {
            ads_printf("Error retrieving object id.\n");
            myrc = kFalse;
            break;
        }
        // open for read or write?
        //
        // No, maybe we should clone it instead?
        //
        if (acdbOpenEntity(pObj, objId, AcDb::kForRead) != Acad::eOk) {
            ads_printf("Error opening object for read.\n");
            myrc = kFalse;
            break;
        }
        if ((pClone = AcDbEntity::cast(pObj->clone())) == NULL) {
            ads_printf("Error cloning object.\n");
            myrc = kFalse;
            break;
        }
        // append cloned entity to the block table record:
        //
        // ...

        // erase original object, just like normal BLOCK command:
        //
        // No, no, no! This entity is in the block now,
        // not in the database ... or what?
        //
        if (pObj->upgradeOpen() != Acad::eOk) {
            ads_printf("Error upgrading open on original object.\n");
            myrc = kFalse;
            break;
        }
        if (pObj->erase() != Acad::eOk) {
            ads_printf("Error erasing original object.\n");
            myrc = kFalse;
            break;
        }
        if (pObj->close() != Acad::eOk) {
            acrx_abort("Error closing original object.\n");
        }
        if (pClone->close() != Acad::eOk) {
            acrx_abort("Error closing cloned object.\n");
        }
    }
    if (ads_ssfree(ss) != RTNORM) {
        ads_printf("Error freeing selection set.\n");
        return myrc;
    }
    if (pBlockTableRec->close() != Acad::eOk) {
        acrx_abort("Error closing block table record.\n");
    }
    return myrc;
}

// listBlock - list block definitions:
//
// The user has the option of listing:
//  - the "model space block"
//  - the "paper space block"
//  - a block defined by name
//  - a block defined by picking a block insertion
//
Boolean
listBlock()
{
    char kw[20];
    char buf[50];
    char *bName = buf;

    ads_name en;
    ads_point pt;
    ads_initget(0, "Model Paper Name");
    int rc = ads_entsel("Model space/Paper space/Name/<Pick insertion>: ",
    en, pt);

    switch (rc) {
    case RTCAN:
    case RTNONE:
        ads_printf("Nothing selected.\n");
        return kTrue;
    case RTERROR:
        ads_printf("Error in ads_entsel.\n");
        return kFalse;
    case RTKWORD:
        if (ads_getinput(kw) != RTNORM) {
            ads_printf("Error in ads_getinput.\n");
            return kFalse;
        }
        if (streq(kw, "Model")) {
            strcpy(bName, ACDB_MODEL_SPACE);
        } else if (streq(kw, "Paper")) {
            strcpy(bName, ACDB_PAPER_SPACE);
        } else {
            rc = ads_getstring(kFalse, "Enter block name: ", bName);
            if (rc!=RTNORM || *bName == '\0') {
                ads_printf("Invalid block name '%s'.\n", bName);
                return kFalse;
            }
        }
        break;

    case RTNORM:
        // retrieve name from block insertion picked:
        //
        AcDbObjectId objId;
        if (acdbGetObjectId(objId, en) != Acad::eOk) {
            ads_printf("Error retrieving object id.\n");
            return kFalse;
        }

        AcDbBlockReference *pInsertObj;
        if (acdbOpenEntity(pInsertObj, objId, AcDb::kForRead) != Acad::eOk) {
            ads_printf("Error opening block insertion for read.\n");
            return kFalse;
        }
        AcDbObjectId bId = pInsertObj->blockTableRecord();
        pInsertObj->close();

        AcDbBlockTableRecord *pBlkDefObj;
        if (acdbOpenEntity(pBlkDefObj, bId, AcDb::kForRead) != Acad::eOk) {
            ads_printf("Error opening block definition for read.\n");
            return kFalse;
        }
        pBlkDefObj->getName(bName);
        pBlkDefObj->close();
        break;
    }

    // at this point, we have obtained the block name,
    // one way or another, and are prepared to iterate
    // over its definition:
    //
    AcDbBlockTable* pBlockTable;
    AcDbBlockTableRecord* pBlockTableRecord;
    AcDbBlockTableRecordIterator* pBlockIterator;

    // open acdbCurDwg's block table for read:
    //
    // ...

    strupr(bName);

    // search for block table record in block table using bName key:
    //
    // ...

    pBlockTable->close();
    AcGePoint3d p = pBlockTableRecord->origin(); // save for display later

    // create interator: start at beginning, 
    // do not skip deleted entites:
    //

    // create a new block interator, assign it to pBlockIterator:
    //
    // ...

    // run iterator through once to get
    // number of entities in block:
    //
    for (long n = 0; !pBlockIterator->done(); pBlockIterator->step()) {
        ++n;
    }
    ads_printf("\nBlock %s at (%.2lf %.2lf %.2lf) with %ld entities:\n==",
    bName, p.x, p.y, p.z, n);
    printObjectInfo(pBlockTableRecord);

    // close the record now that we are done with it
    Acad::ErrorStatus es = pBlockTableRecord->close();
    if (es != Acad::eOk) {
        acrx_abort("Error %d closing block table record.\n", es);
    }

    for (pBlockIterator->start(kTrue, kFalse); !pBlockIterator->done(); 
        pBlockIterator->step()) 
    {
        AcDbEntity *pEntity;

        // retrieve entity pointer from block iterator:
        //
        // ...

        printObjectInfo(pEntity);
        pEntity->close();
    }
    delete pBlockIterator;
    ads_printf("\n");
    return kTrue;
}

// printObjectInfo - utility function to print object info:
//
void
printObjectInfo(const AcDbObjectId& id)
{
    AcDbObject* obj;
    if (acdbOpenObject(obj, id, AcDb::kForRead) != Acad::eOk) {
        ads_printf("printObjectInfo: open object failed.\n");
    } else {
        printObjectInfo(obj);
        obj->close();
    }
}

// printObjectInfo - utility function to print object info:
//
void
printObjectInfo(const AcDbObject* obj)
{

    AcDbHandle objHandle;
    obj->getAcDbHandle(objHandle);

    char handleStr[20];
    objHandle.getIntoAsciiBuffer(handleStr);
    const char* cname = obj->isA()->name();

    ads_printf("Object Id %lx, handle %s, class %s.\n",
        obj->objectId(), handleStr, cname);
}

// ----- Rx interface -----

void
initApp()
{
    ads_printf("Rx Hello 18 -- initApp.\n");
    acedRegCmds->addCommand("HELLO18", "HELLO18", "HELLO18",
        ACRX_CMD_MODAL, &hello18);
}

void
unloadApp()
{
    ads_printf("Rx Hello 18 -- unloadApp.\n");
    acedRegCmds->removeGroup("HELLO18");
}

extern "C" AcRx::AppRetCode
acrxEntryPoint(AcRx::AppMsgCode msg, void* pkt)
{
    switch (msg) {
    case AcRx::kInitAppMsg:
        acrxDynamicLinker->unlockApplication(pkt);
        initApp();
        break;
    case AcRx::kUnloadAppMsg:
        unloadApp();
        break;
    case AcRx::kLoadADSMsg:
        ads_printf("\nRx Hello 18 -- kLoadADSMsg\n");
        break;
    case AcRx::kUnloadADSMsg:
        ads_printf("Rx Hello 18 -- kUnloadADSMsg\n");
        break;
    default:
        ads_printf("Rx Hello 18 -- <unknown message %d>\n", msg);
        break;
    }
    return AcRx::kRetOK;
}
