/* $Id: rxhell15.cpp,v 1.4 1995/09/23 21:28:05 artc Exp $ */
/* 
      (C) Copyright 1993-1995 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

//
// rxhell15.cpp - Rx app to keep track of changes to entities
//
// Jeremy Tammik, Autodesk ESC, 94-08-11
//
// This minimal Rx program attaches a reactor to the database
// to receive notification each time an entity changes.
//
// See RxHello16 for a more elegant version. It is cleaner,
// since it one iteration later. It is also more complex, 
// since it displayes the database modifications in a 
// modeless Windows dialogue box. So you might look at
// this file first to get an idea of the basic AutoCAD
// interaction, and then go on to RxHello16 if you want 
// to use it in any way.
//
// RxHello17 is better still: one iteration further, and uses
// MFC instead of pure Windows API for the dialogue box.
//
// I assume when a new drawing is loaded, the database reactor
// needs to be reattached; therefore, we also attach a 
// dwgLoadReactor to the acedEditor. This can then reattach
// the database reactor whenever needed. Maybe a more reliable
// method for doing this would be using the AcRx::kLoadADSMsg?
// That's what I've ended up doing for the moment ...
//
#include <stdio.h>
#include <stdlib.h>  // for free() calls
#include <string.h>
#include <rxobject.h>
#include <rxregsvc.h>
#include <rxdlinkr.h>
#include <aced.h>
#include <dbents.h>
#include <dbsymtb.h>
#include <adslib.h>
#include "dbmr.h" // DbModReactor
#include "dwgr.h" // DwgLoadReactor
#include "rxhell15.h"
#include "adeskabb.h"

// global reactor objects:
//
// They need to be global, so that they can be 
// deleted if the application is terminated.
//
DwgLoadReactor *dwgr = NULL;
DbModReactor *dbmr = NULL;
Boolean dbmrOn; // are we currently watching dwg modifications?

// hello15 - AutoCAD command to watch database modifications:
//
// We offer options switch on and off the database 
// modification reactor, which sets a global 
// boolean switch and calls attachDbmr to do the 
// dirty deed.
//
void
hello15()
{
    dbmrOn = getYorN("Start watching entity modifications");
    attachDbmr();
}

void
attachDbmr()
{
    AcDbDatabase *pDb = acdbCurDwg();

    if(dbmrOn) {
        // 1. if reactor does not already exist create one using dbmr var
        // 2. add it to the current database
        // 3. notify user
        // ...
    } else if(dbmr) {
        // 1. delete it from database
        // 2. delete it from memory
        // 3. reset it to NULL
        // 4. notify user
        // ...
    } else {
        ads_printf("DbModReactor pointer was NULL - not removed.\n");
    }
}

// ----- DbModReactor -----

void
DbModReactor::objectAppended(const AcDbDatabase* dwg,
    const AcDbObject* dbObj)
{
    objDisplay("appended", dbObj);
}

void
DbModReactor::objectModified(const AcDbDatabase* dwg,
    const AcDbObject* dbObj)
{
    objDisplay("modified", dbObj);
}

void
DbModReactor::objectErased(const AcDbDatabase* dwg,
    const AcDbObject* dbObj, Adesk::Boolean pErased)
{
    char buf[40];
    sprintf(buf, "%serased", (pErased ? "" : "not "));
    objDisplay(buf, dbObj);
}

void
DbModReactor::objDisplay(const char* what, const AcDbObject* dbObj) const
{
    AcDbObjectId id = dbObj->objectId();

    AcDbHandle h;
    dbObj->getAcDbHandle(h);

    char hstr[20];
    h.getIntoAsciiBuffer(hstr);
    const char *cname = dbObj->isA()->name();

    ads_printf("\nDbModReactor: obj %s: class %s, id %lx, handle %s.\n",
        what, cname, id, hstr);
}

// ----- DwgLoadReactor -----

void
DwgLoadReactor::dwgFileOpened(AcDbDatabase* pDwg, char* fileName)
{
// There appears to be a bug in ARX that causes AcDbDatabase::headerVar()
// to return Acad::eKeyNotFound instead of Acad::eOk even though it has
// successfully set the rb.resval.rstring to the DWGNAME variable's value.
// So,  for now I'm setting the test to look for either eOk or eKeyNotFound

    resbuf rb;
    Acad::ErrorStatus es = pDwg->getHeaderVar(&rb, "dwgname");
    if (es == Acad::eOk || es == Acad::eKeyNotFound)
        ads_printf("DwgLoadReactor: drawing %s was loaded.\n",
            rb.resval.rstring);
    else
        ads_printf ("Unable to obtain drawing name in openComplete call\n");
}

// getYorN - utility function to get boolean from user:
//
Boolean
getYorN(const char* pStr)
{
    char yorn_str[132];

    yorn_str[0] = 'Y';
    yorn_str[1] = '\0';
    ads_printf("\n%s", pStr);
    ads_initget(0, "No Yes");
    ads_getkword(" -- No/<Yes>:  ", yorn_str);
    return(yorn_str[0] != 'N');
}

void
initApp()
{
    ads_printf("Rx Hello 15 -- initApp.\n");

    dwgr = new DwgLoadReactor();
    acedEditor->addReactor(dwgr);
    ads_printf("Added DwgLoadReactor to acedEditor.\n");

    acedRegCmds->addCommand("HELLO15", "HELLO15", "HELLO15",
        ACRX_CMD_MODAL, &hello15);
}

void
unloadApp()
{
    ads_printf("Rx Hello 15 -- unloadApp.\n");
    delete dwgr;
    dwgr = NULL;
    dbmrOn = kFalse;
    attachDbmr();                         // detach and delete dbmr
    acedRegCmds->removeGroup("HELLO15");
}

extern "C" AcRx::AppRetCode
acrxEntryPoint(AcRx::AppMsgCode msg, void* pkt)
{
    switch (msg) {
    case AcRx::kInitAppMsg:
        acrxDynamicLinker->unlockApplication(pkt);
        initApp();
        break;
    case AcRx::kUnloadAppMsg:
        unloadApp();
        break;
    case AcRx::kLoadADSMsg:
        ads_printf("\nRx Hello 15 -- kLoadADSMsg\n");
        
        // this is called if a NEW drawing is started
        //
        attachDbmr(); 
        break;
    case AcRx::kUnloadADSMsg:
        ads_printf("Rx Hello 15 -- kUnloadADSMsg\n");
        break;
    default:
        ads_printf("Rx Hello 15 -- <unknown message %d>\n", msg);
        break;
    }
    return AcRx::kRetOK;
}


