/* $Id: jblob.cpp,v 1.8 1995/09/23 21:41:16 artc Exp $ */
/* 
      (C) Copyright 1993-1995 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

//
// jblob.cpp - jblob entity
//
// Jeremy Tammik, Autodesk ESC, 94-08-10
//
// Based on bcurve.h and bcurve.cpp by 
// Peter Janson, 10. May 1994.
//
// Added support to draw the entity,
// read and write to DWG, write to DXF,
// explode, list, osnap (near, end, mid,
// cen, quad), grip editing (move, stretch),
// transformations like scale and move.
//
#include <acdb.h>
#include <math.h>
#include <string.h>
#include "adslib.h"
#include "rxobject.h"
#include "rxregsvc.h"
#include "jblob.h"

const char* kClassName = "Jblob";

// define a DXF class body for the entity using the
// appropriate ACRX boilerplate macro:
//
// ... The following line completes ...
ACRX_DXF_DEFINE_MEMBERS(Jblob, AcDbEntity, 0, JBLOB, RxHello9);

Jblob::Jblob() : mnormal(0,0,1)
{
}

Jblob::~Jblob()
{
}

Adesk::Boolean
Jblob::worldDraw(AcGiWorldDraw* wd)
{
    AcGePoint3d line[2];

    // ask for circle to be filled:
    //
    wd->subEntityTraits().setFillType(kAcGiFillAsACAD);
    line[0] = mp;
    line[1] = mpblob;
    wd->geometry().polyline(2, line);
    wd->geometry().circle(mpblob, mrblob, mnormal);
    return Adesk::kTrue;
}

Acad::ErrorStatus
Jblob::dwgInFields(AcDbDwgFiler* filer)
{
    assertWriteEnabled();

    AcDbEntity::dwgInFields(filer);

    // read in the start point,
    // blob point, blob radius
    // and normal vector
    //
    // ... The following 4 lines complete ...

    filer->readItem(&mp);
    filer->readItem(&mpblob);
    filer->readItem(&mrblob);
    filer->readItem(&mnormal);

    return filer->filerStatus();
}

Acad::ErrorStatus
Jblob::dwgOutFields(AcDbDwgFiler* filer) const
{
    assertReadEnabled();

    AcDbEntity::dwgOutFields(filer);

    // write out the start point,
    // blob point, blob radius
    // and normal vector
    //
    // ... The following 4 lines complete...

    filer->writeItem(mp);
    filer->writeItem(mpblob);
    filer->writeItem(mrblob);
    filer->writeItem(mnormal);

    return filer->filerStatus();
}

Acad::ErrorStatus
Jblob::dxfInFields(AcDbDxfFiler* filer)
{
    assertWriteEnabled();

    struct resbuf rb;

    Acad::ErrorStatus es = AcDbEntity::dxfInFields(filer);

    if (es != Acad::eOk) {
        return es;
    }

    // check that we are at the correct subclass data:
    //
    // ... The following 3 lines complete ...

    if (!filer->atSubclassData(kClassName)) {
        return Acad::eBadDxfSequence;
    }

    // set the normal's default value:
    //
    // ... The following line completes ...

    mnormal = AcGeVector3d(0, 0, 1); // set default value:

    // loop as long as you can read in items
    // from the filer. Depending on the group
    // code read, set the start point,
    // blob point, blob radius and normal 
    // vector:
    //
    // ... The following 20 lines complete ...

    while (es == Acad::eOk) {
        if ((es = filer->readItem(&rb)) == Acad::eOk) {
            switch(rb.restype) {
            case AcDb::kDxfXCoord:
                mp.set(rb.resval.rpoint[X], rb.resval.rpoint[Y],
                    rb.resval.rpoint[Z]);
                break;
            case AcDb::kDxfXCoord+1:
                mpblob.set(rb.resval.rpoint[X], rb.resval.rpoint[Y],
                    rb.resval.rpoint[Z]);
                break;
            case AcDb::kDxfReal:
                mrblob = rb.resval.rreal;
                break;
            case AcDb::kDxfNormalX:
                mnormal.set(rb.resval.rpoint[X], rb.resval.rpoint[Y],
                    rb.resval.rpoint[Z]);
            }
        }
    }
    return filer->filerStatus();
}

Acad::ErrorStatus
Jblob::dxfOutFields(AcDbDxfFiler* filer) const
{
    assertReadEnabled();

    AcDbEntity::dxfOutFields(filer);

    // write out the subclass name,
    // start point, blob point, blob radius
    // and normal vector.
    //
    // You may want to check 
    // filer->includesDefaultValues()
    // and mnormal != AcGeVector3d(0,0,1)
    // before writing out the normal vector.
    //
    // ... The following 9 lines complete ...

    filer->writeItem(AcDb::kDxfSubclass, kClassName);
    filer->writeItem(AcDb::kDxfXCoord, mp);
    filer->writeItem(AcDb::kDxfXCoord + 1, mpblob);
    filer->writeItem(AcDb::kDxfReal, mrblob);
    if (filer->includesDefaultValues()
        || mnormal != AcGeVector3d(0,0,1))
    {
        filer->writeItem(AcDb::kDxfNormalX, mnormal);
    }
    return filer->filerStatus();
}

Acad::ErrorStatus
Jblob::explode(AcDbVoidPtrArray& entities) const
{
    assertReadEnabled();

    // explode the Jblob by creating 
    // an appropriately positioned
    // line and circle:
    //
    // ... The following 11 lines complete this ...

    AcDbLine *pLine = new AcDbLine;
    pLine->setStartPoint(mp);
    pLine->setEndPoint(mpblob);
    pLine->setNormal(mnormal);
    if (pLine) { entities.append(pLine); }

    AcDbCircle *pCircle = new AcDbCircle;
    pCircle->setCenter(mpblob);
    pCircle->setRadius(mrblob);
    pCircle->setNormal(mnormal);
    if (pCircle) { entities.append(pCircle); }

    return Acad::eOk;
}

void
Jblob::list() const
{
    assertReadEnabled();

    AcDbEntity::list();

    // list your Jblob data with ads_printf:
    //
    // ... The following 6 lines complete this...

    ads_printf("%19sPOINT\n"
        "%16sat point, X=%9.16q0 Y=%9.16q0 Z=%9.16q0\n",
        "", "", mp[X], mp[Y], mp[Z]);
    ads_printf("%19sBLOB\n"
        "%16sat point, X=%9.16q0 Y=%9.16q0 Z=%9.16q0 R=%9.16q0\n",
        "", "", mpblob[X], mpblob[Y], mpblob[Z], mrblob);
}

Acad::ErrorStatus
Jblob::getOsnapPoints(
    AcDb::OsnapMode   osnapMode,
    int         gsSelectionMark,
    const AcGePoint3d&  pickPoint,
    const AcGePoint3d&  lastPoint,
    const AcGeMatrix3d& viewXform,
    AcGePoint3dArray&    snapPoints,
    AcDbIntArray&   geomIds) const
{
    assertReadEnabled();

    switch(osnapMode) {

    case AcDb::kOsModeEnd:
        return osnapEnd(pickPoint, snapPoints);
    case AcDb::kOsModeNear:
        return osnapNear(pickPoint, snapPoints);
    case AcDb::kOsModeMid:
        return osnapMid(pickPoint, snapPoints);
    case AcDb::kOsModeCen:
        return osnapCen(pickPoint, snapPoints);
    case AcDb::kOsModeQuad:
        return osnapQuad(pickPoint, snapPoints);
    case AcDb::kOsModeNode:
    case AcDb::kOsModeIns:
    case AcDb::kOsModePerp:
    case AcDb::kOsModeTan:
    default:
        break;
    }
    return Acad::eInvalidInput;
}

Acad::ErrorStatus
Jblob::getGripPoints(
    AcGePoint3dArray& gripPoints,
    AcDbIntArray& osnapMasks,
    AcDbIntArray& geomIds) const
{
    assertReadEnabled();

    // add all the grip points of a line and a circle, i.e. line endpoints
    // and circle center and quad points. Make use of AcGeVector3d features.
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following 9 lines complete this...

    gripPoints.append(mp);
    gripPoints.append(mp + 0.5 * (mpblob - mp));
    gripPoints.append(mpblob);
    AcGeVector3d xoff(mrblob, 0, 0);
    AcGeVector3d yoff(0, mrblob, 0);
    gripPoints.append(mpblob + xoff);
    gripPoints.append(mpblob + yoff);
    gripPoints.append(mpblob - xoff);
    gripPoints.append(mpblob - yoff);

    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::moveGripPointsAt(
    const AcDbIntArray& indices,
    const AcGeVector3d& offset)
{
    AcGePoint3d oldquad, newquad;

    assertWriteEnabled();

    for (int i = 0; i < indices.length(); i++) {
        int idx = indices[i];
        switch(idx) {
        case 0:
            mp += offset;
            continue; // stretch begin point
        case 1:
            mp += offset;
            mpblob += offset;
            continue; // move
        case 2:
            mpblob += offset;
            continue; // stretch blob center

        // stretch blob radius:
        //
        case 3:
            oldquad = mpblob + AcGeVector3d(mrblob, 0, 0);
            break;
        case 4:
            oldquad = mpblob + AcGeVector3d(0, mrblob, 0);
            break;
        case 5:
            oldquad = mpblob - AcGeVector3d(mrblob, 0, 0);
            break;
        case 6:
            oldquad = mpblob - AcGeVector3d(0, mrblob, 0);
            break;
        }
        newquad = oldquad + offset;
        mrblob = newquad.distanceTo(mpblob);

        // update other grip points as well?
    }
    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::getStretchPoints(AcGePoint3dArray& stretchPoints) const
{
    AcDbIntArray osnapModes;
    AcDbIntArray geomIds;

    assertReadEnabled();

    //  getStretchPoints == getGripPoints
    //
    return getGripPoints(stretchPoints, osnapModes, geomIds);
}

Acad::ErrorStatus
Jblob::moveStretchPointsAt(const AcDbIntArray& indices,
    const AcGeVector3d& offset)
{
    assertWriteEnabled();

    //  moveStretchPointsAt == moveGripPointsAt
    //
    return moveGripPointsAt(indices, offset);
}

Acad::ErrorStatus
Jblob::transformBy(const AcGeMatrix3d& xform)
{
    assertWriteEnabled();

    AcGePoint3d xquad = mpblob + AcGeVector3d(mrblob, 0, 0);
    xquad.transformBy(xform);
    mp.transformBy(xform);
    mpblob.transformBy(xform);
    mrblob = mpblob.distanceTo(xquad);

    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::osnapNear(const AcGePoint3d& pickPoint,
    AcGePoint3dArray& snapPoints) const
{
    // for osnap near, define the same points 
    // as the grip points:
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following 9 lines complete this...

    snapPoints.append(mp);
    snapPoints.append(mp + 0.5 * (mpblob - mp));
    snapPoints.append(mpblob);
    AcGeVector3d xoff(mrblob, 0, 0);
    AcGeVector3d yoff(0, mrblob, 0);
    snapPoints.append(mpblob + xoff);
    snapPoints.append(mpblob + yoff);
    snapPoints.append(mpblob - xoff);
    snapPoints.append(mpblob - yoff);
    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::osnapEnd(const AcGePoint3d& pickPoint,
    AcGePoint3dArray& snapPoints) const
{
    // osnap end to the two endpoints
    // of the line:
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following 2 lines complete this...

    snapPoints.append(mp);
    snapPoints.append(mpblob);

    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::osnapMid(const AcGePoint3d& pickPoint,
    AcGePoint3dArray& snapPoints) const
{
    // osnap mid to the line midpoint:
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following line completes this...

    snapPoints.append(mp + 0.5 * (mpblob - mp));

    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::osnapCen(const AcGePoint3d& pickPoint,
    AcGePoint3dArray& snapPoints) const
{
    // osnap center to the circle's center
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following line completes this...

    snapPoints.append(mpblob);

    return Acad::eOk;
}

Acad::ErrorStatus
Jblob::osnapQuad(const AcGePoint3d& pickPoint,
    AcGePoint3dArray& snapPoints) const
{
    // osnap quad to the circle's quad points:
    //
    // NOTE:  The AcGePoint3dArray passed into this function may not be
    // empty,  so be sure to append points to it instead of assigning values
    // to existing elements.
    //
    // ... The following 6 lines complete this...

    AcGeVector3d xoff(mrblob, 0, 0);
    AcGeVector3d yoff(0, mrblob, 0);
    snapPoints.append(mpblob + xoff);
    snapPoints.append(mpblob + yoff);
    snapPoints.append(mpblob - xoff);
    snapPoints.append(mpblob - yoff);

    return Acad::eOk;
}
