/* $Id: rxhello5.cpp,v 1.1 1994/10/18 16:31:18 edwardb Exp $ */
/* 
      (C) Copyright 1993-1994 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

/*
	rxhello5.cpp

	The minimal Rx program defining a protocol extension.
	It says 'hello' both at initModule and syncModule.
	It defines a class HelloPex with a method
	HelloPex::hello_pex(), which defines an extension
	protocol for the HelloClass defined by RxHello4 and
	which appropriately enough also says hello.

	We have the three modules server, api library and client.

	The client application defines an AutoCAD command HELLO5
	which invokes the protocol extension function
	HelloClass::hello_pex through the Rx protocol extension
	mechanisms, which actually lead up to HelloPex::hello_pex,
	of course.

	To extend the protocol of HelloClass, we first check that
	this class is available, i.e. the RxHello4 services have
	been registered. Therefore, we do not attach the protocol
	extension in initModule, but in syncModule, after the check.

	Jeremy Tammik, Autodesk ESC, 93-11-22
*/
#include <rxobject.h>
#include <rxregsvc.h> // register services
#include <adslib.h>
#include "../rxhell04/hello4.h"
#include "helloimp.h"

// syncModule -- actually, called by initModule for the time being:

void syncModule ()
{
	int rx4_ok = 0;

	ads_printf ("Rx Hello 5 Server -- syncModule\n");

	rx4_ok = acrxServiceIsRegistered (RX4_SERVICE);
	ads_printf ("%s has %sbeen registered.\n", RX4_SERVICE,
		rx4_ok ? "" : "NOT ");

	if (rx4_ok) {
		HelloClass::desc()->addX (HelloPex::desc(), newHelloPex());
	}
	else {
		ads_printf ("* No protocol extension defined for HelloClass! *\n");
	}
}

void initModule()
{
	ads_printf ("Rx Hello 5 Server -- initModule\n");

	/* Step 1:

	   Add classes to the runtime-identifiable class hierarchy.

	   This call was defined by use of the ACRX_CLASS() macro on
	   HelloPex and registers its class objects with Rx.
	*/
	initHelloPex();
	initHelloPexImp();

	/* Step 2:

	   Build class relationships before proceeding.	 Use after
	   making a set of init<Classname>() calls which add a contiguous
	   growth onto the inheritance tree.
	*/
	acrxBuildClassHierarchy();

	/* Step 3:

	   Do any other initialization you need.  The point is that it is
	   best to do the classes first, since the rest of the initialization
	   is usually concerned with them.

	   This example application adds a new object to the system dictionary
	   and registers a service name, needed for symbolic function
	   export,	and a handy object for verifying the presence of this
	   application.

	*/
	acrxSysRegistry()->atPut ("HelloP_Extension", new HelloPexImp());

	acrxRegisterService (RX5_SERVICE);

	// since we get no syncModule call, let's do it ourselves
	// at this point -- not very elegant, we rely on RxHello4
	// having been loaded before us, otherwise this will not
	// work -- at least it will warn the user, though:
	//
	syncModule ();
}

extern "C" AcRx::AppRetCode acrxEntryPoint (AcRx::AppMsgCode msg, void*)
{
	switch(msg) {
	case AcRx::kInitAppMsg:
		initModule ();
		break;
	case AcRx::kUnloadAppMsg:
		ads_printf ("Rx Hello 5 Server -- kUnloadAppMsg\n");
		break;
	case AcRx::kLoadADSMsg:
		ads_printf ("Rx Hello 5 Server -- kLoadADSMsg\n");
		break;
	case AcRx::kUnloadADSMsg:
		ads_printf ("Rx Hello 5 Server -- kUnloadADSMsg\n");
		break;
	default:
		ads_printf ("Rx Hello 5 Server -- <unknown message %d>\n", msg);
		break;
	}
	return AcRx::kRetOK;
}
