/* $Id: rxhell10.cpp,v 1.3 1995/08/16 17:22:48 elevy Exp $ */
/* 
      (C) Copyright 1993-1995 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

//
// rxhell10.cpp - Rx app to keep track of changes to entities
//
// Jeremy Tammik, Autodesk ESC, 94-08-15
//
// RxHello10 creates, lists and benchmarks R13 entity groups.
//
// Beside the command HELLO10 defined in this module, 
// this application also defines all the commands provided 
// by Bill's grouptst.cpp which is simply linked in as is.
//
// Benchmark results in r074:
//
// Create 100 groups
// Average 10 entities per group
// Search for random groups and create selections sets of
//   their entities 1000 times -- produces the results:
// 100 groups with 455 entities created in 4.45 seconds.
// 1000 searches for groups retrieving 4174 entities performed 
//   in 10.92 seconds.
//
// Beside the command HELLO10 defined in this module, 
// this application also defines all the commands provided 
// by Bill's grouptst.cpp which is simply linked in as is.
//
// Benchmark results in r074:
//
// Create 100 groups
// Average 10 entities per group
// Search for random groups and create selections sets of
//   their entities 1000 times -- produces the results:
// 100 groups with 455 entities created in 4.45 seconds.
// 1000 searches for groups retrieving 4174 entities performed 
//   in 10.92 seconds.
//
// First try:
//
// 100 groups with 455 entities created in 4.45 seconds.
// 1000 searches for groups retrieving 4174 entities performed
//   in 10.92 seconds.
//
// Second try:
//
// 100 groups with 473 entities created in 4.05 seconds.
// 1000 searches for groups retrieving 4446 entities performed
//   in 11.23 seconds.
//
// Both the AcDbGroup constructor and the AcDbDictionary
// getAt and setAt methods could do with some more 
// const-ness ... the key should be 'const char * const'.
//
// There is a rather subtle bug either in my createGroup
// function, or in the AcDb group library ... when I use
// Bill's CREATEGROUP and APPENDGROUP commands to create
// a group with one single entity and then list it using
// either his LISTGROUP or my HELLO10 List command, the
// entity shows up. If I create an identical group with 
// my HELLO10 Create command, the group reports 1 entity
// in nbrEntites, but the iterator does not list the 
// single entity the first time either LISTGROUP or 
// HELLO10 List is called ... the senond and later calls
// list it perfectly, though. The only difference I can 
// see is that using CREATEGROUP and APPENDGROUP commands 
// individuall closes and reopens the group between the 
// creation and appending. Have I made a mistake in my
// code, or is there a buggie hidden in there somewhere?
//
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <aced.h>
#include <dbsymtb.h>
#include <dbents.h>
#include <dbgroup.h>
#include <adslib.h>
#include "rxhell10.h"
#include "adeskabb.h"

#define streq(s,t) (strcmp((s),(t))==0)

// hello10 - AutoCAD command to manipulate entity groups:
//
// Create, list and benchmark R13 entity groups.
//
void hello10( void )
{
	int rc;
	char kw[20];
	AcDbObjectId dictId;

	ads_printf( "This is hello10() in RxHello10 ...\n" );
	ads_initget( 0, "Create List Benchmark" );
	rc = ads_getkword( "Benchmark/Create/<List>: ", kw );

	if( rc == RTNONE ) {
		strcpy( kw, "List" );
		rc = RTNORM;
	}
	if( rc != RTNORM ) {
		ads_printf( "Nothing selected.\n" );
		return;
	}
	if( streq( kw, "List" ) ) {
		listGroup();
	}
	else if( streq( kw, "Create" ) ) {
		createGroup();
	}
	else if( streq( kw, "Benchmark" ) ) {
		bmGroup();
	}
	return;
}

// createGroup - create a group using selected entites:
//
Boolean createGroup()
{
	char gName[80], gDesc[80];
	ads_name ss;
	long len;
	int rc;
	//
	// get name and description of group:
	//
	rc = ads_getstring( kFalse, "Enter group name: ", gName );
	if( rc==RTCAN ) { return kTrue; }
	if( rc!=RTNORM || *gName == '\0' ) {
		ads_printf( "Invalid group name '%s'.\n", gName );
		return kFalse;
	}
	rc = ads_getstring( kTrue, "Enter group description:", gDesc);
	if( rc==RTCAN ) { return kTrue; }
	if( rc!=RTNORM ) {
		ads_printf( "Invalid group description '%s'.\n", gDesc );
		return kFalse;
	}
	//
	// get entities for group:
	//
	rc = ads_ssget( NULL, NULL, NULL, NULL, ss );
	if ( rc!=RTNORM ) {
		ads_printf( "Nothing selected.\n" );
		return kFalse;
	}
	rc = ads_sslength( ss, &len );
	if ( rc!=RTNORM ) {
		ads_printf( "Invalid selection set.\n" );
		ads_ssfree( ss );
		return kFalse;
	}
	//
	// create and name a new group table record:
	//
	AcDbGroup* grp = new AcDbGroup( gName, gDesc, Adesk::kTrue );
	//
	// add the group to the group dictionary:
	//
	Acad::ErrorStatus es;
	AcDbDictionary* grpDict;
	es = acdbCurDwg()->groupTable( grpDict, AcDb::kForWrite );
    if ( es != Acad::eOk ) {
		ads_printf( "Error %d opening group table for write.\n", es );
		ads_ssfree( ss );
		// delete AcDbGroup grp here?
		return kFalse;
	}
	AcDbObjectId grpId;
	es = grpDict->setAt( gName, (AcDbObject*)grp, grpId );
	if( es != Acad::eOk ) {
		ads_printf("Error %d adding group to dictionary.\n", es );
		grpDict->close();
		ads_ssfree( ss );
		// delete AcDbGroup grp here?
		return kFalse;
	}
	grpDict->close();
	//
	// add entities to the group record:
	//
	ads_name en;
	AcDbObjectId eId;
	Boolean myrc = kTrue;
	for( long i=0; i<len; ++i ) {
		ads_ssname( ss, i, en );
		if( acdbGetObjectId( eId, en ) != Acad::eOk ) {
			ads_printf( "Error retrieving object id.\n" );
			myrc = kFalse;
			break;
		}
		es = grp->append( eId );
		if( es != Acad::eOk ) {
			ads_printf( "Error %d appending object to group.\n" );
			myrc = kFalse;
			break;
		}
	}
	if( ads_ssfree( ss ) != RTNORM ) {
		ads_printf( "Error freeing selection set.\n" );
		return myrc;
	}
	if( grp->close() != Acad::eOk ) {
		ads_printf( "Error closing group table record.\n" );
		return myrc;
	}
	return myrc;
}

// listGroup - list group definitions:
//
// Strange ... if I create a group with exactly one 
// element, nbrEntities returns 1 correctly all the 
// time, but the AcDbGroupIterator returns notDone=0
// immeiately the first time it is created. In all 
// further calls, it loops thorugh 1 entity as well.
//
Boolean listGroup()
{
	int rc;
	char gName[50];
	Acad::ErrorStatus es;
	AcDbDictionary * grpDict;
	AcDbGroup * grp;

	rc = ads_getstring( kFalse, "Enter group name: ", 
		gName );
	if ( rc!=RTNORM || *gName == '\0' ) {
    	ads_printf( "Invalid group name '%s'.\n", gName );
        return kFalse;
	}
	es = acdbCurDwg()->groupTable( grpDict, AcDb::kForRead );
	if( es != Acad::eOk) {
		ads_printf( "Error %d opening group table.\n", es );
		return kFalse;
	}
	// see if needed ... nope, names are case sensitive:
	//strupr( gName ); 
	AcDbObject *obj;
	es = grpDict->getAt( gName, AcDb::kForRead, obj );
	grp = (AcDbGroup*) obj;
	if (es != Acad::eOk) {
		ads_printf("Error %d opening group '%s'.\n", es, gName );
	 	grpDict->close();
		return kFalse;
	}
 	grpDict->close();
	//
	// create interator: start at beginning, 
	// do not skip deleted entites:
	//
	AcDbGroupIterator* i = grp->newIterator();
	if ( i == NULL ) {
		ads_printf( "Error creating group iterator.\n" );
		return kFalse;
	}
	//
	// read group attributes:
	//
	ads_printf( "\nGroup %s, desc '%s', "
		"%sselectable, %ld entities:\n==",
		gName, grp->description(),
		(grp->isSelectable() ? "" : "not "), 
		grp->numEntities() );
	printObjectInfo( grp );

	while( i->notDone() ) {
		i->nextElement();
		AcDbObjectId id = i->objectId();
		printObjectInfo( id );
	}
	delete i;
	ads_printf("\n");
	es = grp->close();
	if (es != Acad::eOk) {
		ads_printf( "Error %d closing group table record.\n", es );
		return kFalse;
	}
	return kTrue;
}

// bmGroup - list group definitions:
//
// Hmm, how shall we do this? I propose to
// ask the user for a number N, then create 
// N groups of entities. Each group will contain
// a random number [1..M] of entities, which will 
// be lines, randomly positioned in the square
// [0,10]x[0,10]. 
//
// First I'll measure the time to create all the 
// groups, then to perform a certain number of 
// searches.
//
// I should benchmark two kinds of searches:
//  - given a group name, return all the entities
//  - given an entity, return the group(s) it belongs
//    to -- I don't know how to do this at the moment,
//    save by adding an xdata pointer to each entity;
//    but there are probably more elegant and efficient 
//    methods in the grand land of Rx, no?
// 
Boolean bmGroup()
{
	int rc;
	int nGroups;
	int maxSize; // max number of entities in each group
	int nEnts; // number of entities in one group
	int nSearch; // number of times to search for a group
	long totalEnts = 0; // total number entities in all groups
	AcGePoint3d p, q; // start and endpoint of line

	rc = ads_getint( "Number of groups <1000>: ", &nGroups );
	if( rc==RTNONE ) { nGroups = 1000; rc = RTNORM; }
	if( rc==RTCAN ) { return kTrue; }
	if( rc!=RTNORM ) { return kFalse; }

	rc = ads_getint( "Max number of entities per group <100>: ", 
		&maxSize );
	if( rc==RTNONE ) { maxSize = 100; rc = RTNORM; }
	if( rc==RTCAN ) { return kTrue; }
	if( rc!=RTNORM ) { return kFalse; }

	rc = ads_getint( "Number of times to search for groups <1000>: ", 
		&nSearch );
	if( rc==RTNONE ) { nSearch = 1000; rc = RTNORM; }
	if( rc==RTCAN ) { return kTrue; }
	if( rc!=RTNORM ) { return kFalse; }

	if( getYorN( "Random seed" ) ) {
		// seed the random-number generator 
		// with current time so that the numbers 
		// will be different every time we run:
		//
		srand( (unsigned)time( NULL ) );
	}
	ads_printf( "Creating groups:\n" );
	int i;
	char gName[20];
	AcDbGroup* grp;
	AcDbObjectId eId;
	Boolean myrc = kTrue;

	// start timer:
	ads_real timer1 = now();
	for( i=0; i<nGroups && myrc; ++i ) {
		if( ads_usrbrk() ) {
			ads_printf( "\nInterrupted creating group %d.\n", i );
			myrc = kFalse;
			break;
		}
		ads_printf( "\r%d", i );
		sprintf( gName, "_JT_RXHELLO_%d", i );
		if( !cGroup( grp, gName ) ) {
			ads_printf( "\nError creating group %d.\n", i );
			myrc = kFalse;
			break;
		}
		nEnts = rand() % maxSize;
		totalEnts += nEnts;
		while( nEnts-- ) {
			// randomize p, q:
			p.x = (rand() % 9) + (1.0 / (1.0 + rand()));
			p.y = (rand() % 9) + (1.0 / (1.0 + rand()));
			p.z = 0.0;
			q.x = (rand() % 9) + (1.0 / (1.0 + rand()));
			q.y = (rand() % 9) + (1.0 / (1.0 + rand()));
			q.z = 0.0;
			if( !cEnt( eId, p, q ) ) {
				ads_printf( "\nError creating entity for "
					"group %d.\n", i );
				myrc = kFalse;
				break;
			}
			if( grp->append( eId ) != Acad::eOk ) {
				ads_printf( "\nError appending object to "
					"group %d.\n", i );
				myrc = kFalse;
				break;
			}
		}
		if( grp->close() != Acad::eOk ) {
			ads_printf( "\nError closing group %d.\n", i );
			myrc = kFalse;
			break;
		}
	}
	// end timer:
	ads_printf( "\n%d groups with %ld entities created "
		"in %.2lf seconds.\n", 
		nGroups, totalEnts, later( timer1 ) );

	// now to search for groups:
	//
	int gNum;
	long len;
	ads_name ss;

	totalEnts = 0L;
	timer1 = now();
	for( i=0; i<nSearch && myrc; ++i ) {
		if( ads_usrbrk() ) {
			ads_printf( "\nInterrupted during search %d.\n", i );
			myrc = kFalse;
			break;
		}
		ads_printf( "\r%d", i );
		gNum = rand() % nGroups;
		sprintf( gName, "_JT_RXHELLO_%d", gNum );
		if( !fGroup( grp, gName ) ) {
			ads_printf( "\nError during search %d for group %s.\n", 
				i, gName );
			myrc = kFalse;
			break;
		}
		// how to find the entities? Creating a selection
		// set would be a valid test, but maybe unfair to
		// Rx, since there are probably more efficient means
		// to manipulate groups of entites in Rx than that
		// old-fashioned ADS method. Still, lets go and do
		// that to start with:
		//
		if( ads_ssadd( NULL, NULL, ss ) != RTNORM ) {
			ads_printf( "\nError creating selection set "
				"during search %d from group %s.\n", i, gName );
			myrc = kFalse;
			break;
		}
		if( !fEnts( grp, ss ) ) {
			ads_printf( "\nError selecting entities during "
				"search %d from group %s.\n", i, gName );
			myrc = kFalse;
			break;
		}
		if( ads_sslength( ss, &len ) != RTNORM ) {
			ads_printf( "\nError retrieving selection set size "
				"during search %d from group %s.\n", i, gName );
			myrc = kFalse;
			break;
		}
		totalEnts += len;
		if( ads_ssfree( ss ) != RTNORM ) {
			ads_printf( "\nError freeing selection set during "
				"search %d from group %s.\n", i, gName );
			myrc = kFalse;
			break;
		}
		if( grp->close() != Acad::eOk ) {
			ads_printf( "\nError closing group %s during search.\n", 
				gName, i );
			myrc = kFalse;
			break;
		}
	}
	ads_printf( "\n%d searches for groups retrieving %ld entities "
		"performed performed in %.2lf seconds.\n", 
		nSearch, totalEnts, later( timer1 ) );

	return kTrue;
}

// cGroup - create a group for benchmarking:
//
// The const's are rejected, since they are not
// set in the AcDbGroup constructor.
//
// Should be 'const char * const gName', but setAt
// won't let me.
//
Boolean cGroup( AcDbGroup*& grp, char * gName )
{
	grp = new AcDbGroup( gName, 
		"Test description", Adesk::kTrue );
	//
	// add the group to the group dictionary:
	//
	Acad::ErrorStatus es;
	AcDbDictionary* grpDict;
	es = acdbCurDwg()->groupTable( grpDict, AcDb::kForWrite );
    if ( es != Acad::eOk ) {
		ads_printf( "Error %d opening group table for write.\n", es );
		// delete AcDbGroup grp here?
		return kFalse;
	}
	AcDbObjectId grpId;
	es = grpDict->setAt( gName, (AcDbObject*)grp, grpId );
	if( es != Acad::eOk ) {
		ads_printf("Error %d adding group to dictionary.\n", es );
		grpDict->close();
		// delete AcDbGroup grp here?
		return kFalse;
	}
	grpDict->close();
	return kTrue;
}

// cEnt - create an entity for benchmarking:
//
// Creates a line from p to q with default values.
// Returns new entity Id for adding to group.
//
Boolean cEnt( AcDbObjectId& eId, AcGePoint3d& p, AcGePoint3d& q )
{
	Boolean myrc = kTrue;
	Acad::ErrorStatus es;
	AcDbDatabase * pDb = acdbCurDwg();
	AcDbLine * dbEnt = new AcDbLine;
	AcDbBlockTable * dbTable;
	AcDbBlockTableRecord * dbRecord;

	dbEnt->setStartPoint( p ); // AcDbLine operations
	dbEnt->setEndPoint( q );
	dbEnt->setDatabaseDefaults( pDb );

	// open block table to search for model space block:
	//
	es = acdbCurDwg()->blockTable( dbTable, AcDb::kForRead );

	if( es != Acad::eOk ) {
		ads_printf( "Unable to open block table.\n" );
		myrc = kFalse;
		goto exit_1;
	}
	// open current record in model space block:
	//
	es = dbTable->getAt( ACDB_MODEL_SPACE, 
		AcDb::kForWrite, dbRecord );

	if( es != Acad::eOk ) {
		ads_printf( "Unable to open model space block.\n" );
		myrc = kFalse;
		goto exit_2;
	}
	// append new entity to model space block:
	//
	es = dbRecord->appendAcDbEntity( eId, dbEnt );

	if( es != Acad::eOk ) {
		ads_printf( "Unable to append entity.\n" );
		myrc = kFalse;
		goto exit_3;
	}
exit_3:
	if( dbRecord->close() != Acad::eOk ) {
		ads_printf( "Unable to close database record.\n" );
		return kFalse;
	}
exit_2:
	if( dbTable->close() != Acad::eOk ) {
		ads_printf( "Unable to close block table.\n" );
		return kFalse;
	}
exit_1:
	if( dbEnt->close() != Acad::eOk ) {
		ads_printf( "Unable to close entity.\n" );
		return kFalse;
	}
	return myrc;
}

// fGroup - find a group for benchmarking:
//
// Should be 'const char * const gName', but getAt
// won't let me.
//
Boolean fGroup( AcDbGroup*& grp, char * gName )
{
	Acad::ErrorStatus es;
	AcDbDictionary* grpDict;

	es = acdbCurDwg()->groupTable( grpDict, AcDb::kForRead );
    if ( es != Acad::eOk ) {
		ads_printf( "Error %d opening group table for read.\n", es );
		return kFalse;
	}
	AcDbObject * obj;
	es = grpDict->getAt( gName, AcDb::kForRead, obj );
	if( es != Acad::eOk ) {
		ads_printf("Error %d retrieving group from dictionary.\n", 
			es );
		grpDict->close();
		return kFalse;
	}
	grp = (AcDbGroup*) obj;
	grpDict->close();
	return kTrue;
}

// fEnts - find a group's entities for benchmarking:
//
// Return a selection set, which the caller must 
// allocate before the call and free afterwards.
//
Boolean fEnts( AcDbGroup * grp, ads_name ss )
{
	Boolean myrc = kTrue;
	AcDbGroupIterator * i = grp->newIterator();
	ads_name en;

	if ( i == NULL ) {
		ads_printf( "Error creating group iterator.\n" );
		return kFalse;
	}
	while( i->notDone() ) {
		i->nextElement();
		i->objectId();
		if( acdbGetAdsName( en, i->objectId() ) != Acad::eOk ) {
			ads_printf( "Error retrieving entity name.\n" );
			myrc = kFalse;
			break;
		}
		if( ads_ssadd( en, ss, ss ) != RTNORM ) {
			ads_printf( "Error adding entity to selection set.\n" );
			myrc = kFalse;
			break;
		}
	}
	delete i;
	return myrc;
}

// printObjectInfo - utility function to print object info:
//
void printObjectInfo( const AcDbObjectId& id )
{
	AcDbObject* obj;

	if ( acdbOpenObject( obj, id, AcDb::kForRead ) != Acad::eOk ) {
		ads_printf( "printObjectInfo: open object failed.\n" );
	}
	else {
		printObjectInfo( obj );
		obj->close();
	}
}

// printObjectInfo - utility function to print object info:
//
void printObjectInfo( const AcDbObject* obj )
{
	AcDbHandle objHandle;
	char handleStr[20];
	char *cname;

	obj->getAcDbHandle( objHandle );
	objHandle.getIntoAsciiBuffer( handleStr );
	cname = obj->objectClass()->getName()->cpyToAscii();

	ads_printf( "Object Id %lx, handle %s, class %s.\n",
		obj->objectId(), handleStr, cname );
	delete cname;
}

// getYorN - utility function to get boolean from user:
//
Boolean getYorN( const char* pStr )
{
	char yorn_str[132];

	yorn_str[0] = 'Y';
	yorn_str[1] = '\0';
	ads_printf( "\n%s", pStr );
	ads_initget( 0, "No Yes" );
	ads_getkword( " -- No/<Yes>:  ", yorn_str );
	return( yorn_str[0] != 'N' );
}

// timer functions:
//
ads_real now()
{
	struct resbuf rb;
	ads_getvar( "DATE", &rb );
	return rb.resval.rreal;
}

ads_real later( ads_real t1 )
{
	struct resbuf rb;
	ads_getvar( "DATE", &rb );
	return 86400.0 * (rb.resval.rreal - t1);
}

// ----- Rx interface -----

void initApp()
{
	ads_printf( "Rx Hello 10 -- initApp.\n" );
	acedRegCmds->addCommand(
		"HELLO10", "HELLO10", "HELLO10", ACRX_CMD_MODAL, &hello10 );
}

void unloadApp()
{
	ads_printf( "Rx Hello 10 -- unloadApp.\n" );
	acedRegCmds->removeGroup( "HELLO10" );
}

extern "C" AcRx::AppRetCode acrxEntryPoint( AcRx::AppMsgCode msg, void* )
{
	switch( msg ) {
	case AcRx::kInitAppMsg:
		initApp();
		initModule(); // register commands in grouptst.cpp
		break;
	case AcRx::kUnloadAppMsg:
		unloadApp();
		break;
	case AcRx::kLoadADSMsg:
		ads_printf( "\nRx Hello 10 -- kLoadADSMsg\n" );
		break;
	case AcRx::kUnloadADSMsg:
		ads_printf( "Rx Hello 10 -- kUnloadADSMsg\n" );
		break;
	default:
		ads_printf( "Rx Hello 10 -- <unknown message %d>\n", msg );
		break;
	}
	return AcRx::kRetOK;
}
