/* $Id: rxhell19.cpp,v 1.2 1995/05/16 18:47:31 jamesr Exp $ */
/* 
      (C) Copyright 1993-1995 by Autodesk, Inc.

******************************************************************************
*                                                                            *
* The information contained herein is confidential, proprietary to Autodesk, *
* Inc., and considered a trade secret as defined in section 499C of the      *
* penal code of the State of California.  Use of this information by anyone  *
* other than authorized employees of Autodesk, Inc. is granted only under a  *
* written non-disclosure agreement, expressly prescribing the scope and      *
* manner of such use.                                                        *
*                                                                            *
******************************************************************************/

//
// rxhell19.cpp - Rx app to keep track of changes to entities
//
// Jeremy Tammik, Autodesk ESC, 94-08-15
//
// RxHello19 creates and manipulates extended entity data 
// and extended dictionaries.
//
// The extended dictionary part has not been implemented yet.
//
// Check out acdbtest.
//
// Xdata is more useful for storing small objects, since there is 
// less overhead than attaching a dictionary, and then having
// the dictionaries AcDbObject object plus the AcDbObject overhead
// for each entry into it.
//
#include <stdio.h>
#include <string.h> // strcpy()
#include <aced.h>
#include <dbsymtb.h>
#include <adslib.h>
#include "rxhell19.h"
#include "adsprint.h" // ostream acout
#include "adeskabb.h"

#define streq(s,t) (strcmp((s),(t))==0)

// hello19 - AutoCAD command to manipulate xdata and 
//           extended dictionaries:
//
// Creates, iterates and modifies block definitions
// in the AutoCAD database.
//
void hello19( void )
{
	int rc;
	char kw[20];
	AcDbObjectId dictId;

	ads_printf( "This is hello19() in RxHello19 ...\n" );
	ads_initget( 0, "List Xdata Dictionary" );
	rc = ads_getkword( "Dictionary add/Xdata add/<List both>: ", kw );

	if( rc == RTNONE ) {
		strcpy( kw, "List" );
		rc = RTNORM;
	}
	if( rc != RTNORM ) {
		ads_printf( "Nothing selected.\n" );
		return;
	}
	if( streq( kw, "List" ) ) { listBoth(); }
	else if( streq( kw, "Dictionary" ) ) { addDict(); }
	else if( streq( kw, "Xdata" ) ) { addXdata(); }
	return;
}

// ObjectPointer19 class definition:
//
// Custom object for inserting into our dictionary.
//
ACRX_CLASS( ObjectPointer19, AcDbObject )

public:
    ObjectPointer19();
    void          setId( const AcDbObjectId& id );
    AcDbObjectId  getId() const;
    virtual void  print() const;

private:
    AcDbObjectId  mId;

END_ACRX_CLASS

ACRX_DXF_CLASS_OBJ_BODY( ObjectPointer19, AcDbObject, 0, OBJPTR19, RxHello19 );

ObjectPointer19::ObjectPointer19() : mId(0) { }

void ObjectPointer19::setId( const AcDbObjectId& pId ) { mId=pId; }

AcDbObjectId ObjectPointer19::getId() const { return mId; }

void ObjectPointer19::print() const {
	assertReadEnabled();
	printObjectInfo( mId );
}

Boolean addDict()
{
	AcDbObjectId eId ;
	AcDbEntity * ent;
	Acad::ErrorStatus es;

	ent = selectEntity( eId, AcDb::kForWrite );

	if( !ent ) { return kFalse; }

	es = ent->createExtensionDictionary();

	if( Acad::eOk != es ) {
		ads_printf( "Error %d creating extension dictionary.\n", es );
		ads_printf( "We'll go ahead and see whether one already exists ...\n" );
	}
	AcDbObjectId edId = ent->extensionDictionary();
	ent->close();

	AcDbDictionary * edict;
	
	es = acdbOpenObject( edict, edId, AcDb::kForWrite );

	if( Acad::eOk != es ) {
		ads_printf( "Error %d opening entity's extension dictionary.\n", es );
		return kFalse;
	}

	// in order not to "pollute" the entity's
	// extension dictionary with too many entries
	// from this application, we insert one single
	// object: another dictionary. Then, we can
	// put as many things as we want into that
	// without disturbing anybody else:
	//
	AcDbObject * obj;
	AcDbDictionary * dict;

	es = edict->getAt( APPID, AcDb::kForWrite, obj );

	if( Acad::eOk == es ) { // convert obj to dict:
		dict = AcDbDictionary::cast( obj );
		if( !dict ) {
			ads_printf( "Error %d casting %s entry on extension dictionary "
				"to AcDbDictionary.\n", es );
			return kFalse;
		}
	}
	else { // if( Acad::eOk != es ) // probably not there, try to create:
		dict = new AcDbDictionary;
		AcDbObjectId dictId;

		// if this dictionary already has an entry with this key,
		// this call causes an access violation:
		//
		es = edict->setAt( APPID, dict, dictId );
		if( Acad::eOk != es ) {
			ads_printf( "Error %d adding application dictionary "
				"to extension dictionary.\n", es );
			return kFalse;
		}
	}
	edict->close(); // can we do this already?

	int rc;
	char objName[132];

	rc = ads_getstring( kTrue, "Enter dictionary entry key: ", objName );

	if ( rc!=RTNORM || *objName == '\0' ) {
		ads_printf( "Invalid dictionary entry name '%s'.\n", objName );
		dict->close();
		return kFalse;
	}
	// create a new test entry ObjectPointer19
	// add it to the dictionary, demonstrating 
	// instantiation and addition of a
	// custom object class into the database.
	//
	ent = selectEntity( "\nSelect another entity: ", eId, AcDb::kForRead );

	if ( ent == NULL ) {
		ads_printf( "Nothing selected.\n" );
		dict->close();
		return kFalse;
	}
	ent->close();  // didn't really want do anything, just get the Id.

	ObjectPointer19 * newObj = new ObjectPointer19;

	newObj->setId( eId );
	newObj->print();

	AcDbObjectId retObjId;

	es = dict->setAt( objName, newObj, retObjId );
	if ( Acad::eOk != es ) {
		ads_printf( "Error %d adding new entry to dictionary.\n", es );
		dict->close();
		return kFalse;
	}
	if( newObj->close() != Acad::eOk ) {
		ads_printf( "Failed to close new object.\n" );
		dict->close();
		return kFalse;
	}
	if( dict->close() != Acad::eOk ) {
		ads_printf( "Failed to close dictionary.\n" );
		return kFalse;
	}
	return kTrue;
}

Boolean addXdata()
{
	AcDbObjectId eId ;
	AcDbEntity * ent;
	Acad::ErrorStatus es;
	struct resbuf * rb;

	ent = selectEntity( eId, AcDb::kForWrite );

	if( !ent ) { return kFalse; }

	if( !regapp( APPID ) ) {
		ads_printf( "Unable to register application id.\n" );
		ent->close();
		return kFalse;
	}
	if ( (rb = ads_buildlist (
		AcDb::kDxfRegAppName, APPID,
        AcDb::kDxfXdInteger32, 25, 0 )) == NULL ) 
	{
		ads_printf( "Error allocating resbufs for xdata.\n" );
		ent->close();
		return kFalse;
	}
	es = ent->setXData( rb );
    if ( es != Acad::eOk ) {
		ads_printf( "Error %d in setXData.\n", es );
		ent->close();
		return kFalse;
	}
	if ( (es = ent->close()) != Acad::eOk ) {
		ads_printf( "Error %d closing entity.\n", es );
		return kFalse;
	}
	//
	// do we free the resbuf memory here again, or
	// has it been 'adopted' by AutoCAD?
	//
	// Let's try freeing it and see what happens:
	//
	int rc = ads_relrb( rb );

	if ( rc != RTNORM ) {
		ads_printf( "Error %d freeing xdata resbuf list.\n", rc );
		return kFalse;
	}
	return kTrue;
}

Boolean listBoth()
{
	AcDbObjectId eId ;
	AcDbEntity * ent;
	Acad::ErrorStatus es;
	struct resbuf * rb;

	ent = selectEntity( eId, AcDb::kForRead );

	if( !ent ) { return kFalse; }

	rb = (struct resbuf *) ent->xData( APPID );

	acout << "\nXdata: " << rb << endl;
	//
	// do we free the resbuf memory here again, or
	// is it still managed by AutoCAD?
	//
	// Let's try freeing it and see what happens:
	//
	// AutoCAD error aborting
	// FATAL ERROR: At allocdbg.c 455 Damaged buffer 0
	// bytes allocated at line 0 in file l
	//
#if 0
	int rc = ads_relrb( rb );

	if ( rc != RTNORM ) {
		ads_printf( "Error %d freeing xdata resbuf list.\n", rc );
		return kFalse;
	}
#endif

	AcDbObjectId edId = ent->extensionDictionary();

	if ( (es = ent->close()) != Acad::eOk ) {
		ads_printf( "Error %d closing entity.\n", es );
		return kFalse;
	}
	AcDbDictionary * edict;

	es = acdbOpenObject( edict, edId, AcDb::kForRead );

	if( Acad::eOk != es ) {
		ads_printf( "Error %d opening entity's extension dictionary.\n", es );
		return kFalse;
	}
	AcDbDictionaryIterator * i = edict->newIterator();

	for (i = edict->newIterator();  !i->done();  i->nextElement()) {
		char* entryName = i->name();
		ads_printf( "Name: %s;  ", entryName );
		delete entryName;
		printObjectInfo( i->objectId() );
	}
	delete i;

	AcDbObjectId dictId;
	AcDbDictionary * dict;

	es = edict->getAt( APPID, dictId );

	if( Acad::eOk != es ) {
		ads_printf( "No RxHello19 dictionary found in entity's extension dictionary.\n" );
		return kFalse;
	}
	es = acdbOpenObject( dict, dictId, AcDb::kForRead );

	if( Acad::eOk != es ) {
		ads_printf( "Error %d opening RxHello19 dictionary in extension dictionary.\n",
			es );
		return kFalse;
	}
	ads_printf( "\n=== Dictionary contents:\n" );
	for( i = dict->newIterator(); !i->done(); i->nextElement()) {
		char * entryName = i->name();
		ads_printf( "Name: %s;  ", entryName );
		delete entryName;
		printObjectInfo( i->objectId() );
	}
	delete i;
	ads_printf( "=== End of dictionary contents.\n" );

	es = edict->close();
	if( Acad::eOk != es ) {
		ads_printf( "Error %d closing extension dictionary.\n", es );
		return kFalse;
	}
	es = dict->close();
	if( Acad::eOk != es ) {
		ads_printf( "Error %d closing RxHello19 dictionary.\n", es );
		return kFalse;
	}
	return kTrue;
}

// regapp - utility function to register application:
//
// Check if registered first; if not, do so.
// Return true if all went well, i.e. either it was 
// already registered, or the registration was successful.
//
Boolean regapp( const char * const appId )
{
	AcDbDatabase *pDb = acdbCurDwg();
	AcDbRegAppTable *pRegAppTable = NULL;
	AcDbRegAppTableRecord *pRecord = NULL;
	AcDbObjectId id;

	if( pDb == NULL ) {
		ads_printf( "regapp(): Unable to access database.\n" );
		return kFalse;
	}
	// ads_snvalid returns RTERROR or 0 or 1:
	//
	int rc = ads_snvalid( appId, kFalse );

	if( !appId || *appId=='\0' || rc!=1/*RTNORM*/ ) {
		ads_printf( "Invalid application id: %s.\n", appId );
		return kFalse;
	}
	if( pDb->regAppTable( pRegAppTable, AcDb::kForRead ) != Acad::eOk ) {
		ads_printf( "Error opening regapp table for read.\n" );
		return kFalse;
	}
	if( pRegAppTable->has( appId ) ) { // application is registered
		pRegAppTable->close();
		return kTrue;
	}
	if( pRegAppTable->upgradeOpen() != Acad::eOk ) {
		ads_printf( "Error upgrading open for regapp table.\n" );
		pRegAppTable->close();
		return kFalse;
	}
	pRecord = new AcDbRegAppTableRecord();

	if( pRecord == NULL ) {
		ads_printf( "Error instantiating new regapp table entry.\n" );
		pRegAppTable->close();
		return kFalse;
	}
	pRecord->setName( appId );

	if( pRegAppTable->add( id, pRecord ) != Acad::eOk ) {
		ads_printf( "Error adding new regapp table entry.\n" );
		pRecord->close();
		// delete record, free up memory somewhere?
		pRegAppTable->close();
		return kFalse;
	}
	if( pRecord->close() != Acad::eOk ) {
		ads_printf( "Error closing new regapp table entry.\n" );
		pRegAppTable->close();
		return kFalse;
	}
	if( pRegAppTable->close() != Acad::eOk ) {
		ads_printf( "Error closing regapp table.\n" );
		return kFalse;
	}
	return kTrue;
}

// selectEntity - utility function to select entity
//
// Select entity, retrieve object Id and open it,
// creating and returning a valid object pointer.
//
AcDbEntity* selectEntity( AcDb::OpenMode openMode )
{
	AcDbObjectId eId;
	return selectEntity( eId, openMode );
}

AcDbEntity* selectEntity( AcDbObjectId& eId, AcDb::OpenMode openMode )
{
	return selectEntity( "\nSelect an entity: ", eId, openMode );
}

AcDbEntity* selectEntity( const char * const prompt, AcDbObjectId& eId, AcDb::OpenMode openMode )
{
	int rc;
	ads_name en;
	ads_point pt;
	Acad::ErrorStatus es;

	rc = ads_entsel( prompt, en, pt );

	if (rc != RTNORM) {
		ads_printf( "Nothing selected.\n", rc );
		return NULL;
	}
	es = acdbGetObjectId( eId, en );

	if ( es != Acad::eOk ) {
		ads_printf( "Either acdbGetObjectId or ads_entsel failed: "
			"Entity name <%lx,%lx>, error %d.\n", en[0], en[1], es );
		return NULL;
	}
	AcDbEntity * ent;

	es = acdbOpenEntity( ent, eId, openMode );
	if ( es != Acad::eOk ) {
		ads_printf( "acdbOpenEntity failed with error %d.\n", es );
		return NULL;
	}
	return ent;
}

// printObjectInfo - utility function to print object info:
//
void printObjectInfo( const AcDbObjectId& id )
{
	AcDbObject* obj;

	if ( acdbOpenObject( obj, id, AcDb::kForRead ) != Acad::eOk ) {
		ads_printf( "printObjectInfo: open object failed.\n" );
	}
	else {
		printObjectInfo( obj );
		obj->close();
	}
}

// printObjectInfo - utility function to print object info:
//
void printObjectInfo( const AcDbObject* obj )
{
	AcDbHandle h;
	char hstr[20];
	char *cstr;

	obj->getAcDbHandle( h );
	h.getIntoAsciiBuffer( hstr );
	cstr = obj->objectClass()->getName()->cpyToAscii();
	ads_printf( "%s, object Id %lx, handle %s.\n",
		cstr, obj->objectId(), hstr );
	delete cstr;
}

// ----- Rx interface -----

void initApp()
{
	ads_printf( "Rx Hello 19 -- initApp.\n" );
	initObjectPointer19();
	acrxBuildClassHierarchy();
	acedRegCmds->addCommand(
		"HELLO19", "HELLO19", "HELLO19", ACRX_CMD_MODAL, &hello19 );
}

void unloadApp()
{
	ads_printf( "Rx Hello 19 -- unloadApp.\n" );
	acedRegCmds->removeGroup( "HELLO19" );
}

extern "C" AcRx::AppRetCode acrxEntryPoint( AcRx::AppMsgCode msg, void* )
{
	switch( msg ) {
	case AcRx::kInitAppMsg:
		initApp();
		break;
	case AcRx::kUnloadAppMsg:
		unloadApp();
		break;
	case AcRx::kLoadADSMsg:
		ads_printf( "\nRx Hello 19 -- kLoadADSMsg\n" );
		break;
	case AcRx::kUnloadADSMsg:
		ads_printf( "Rx Hello 19 -- kUnloadADSMsg\n" );
		break;
	default:
		ads_printf( "Rx Hello 19 -- <unknown message %d>\n", msg );
		break;
	}
	return AcRx::kRetOK;
}
