#ifndef AD_DBHANDLE_H
#define AD_DBHANDLE_H 1

// $Id: dbhandle.h,v 1.15 1995/08/19 22:53:40 jamesr Exp $
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary
// to Autodesk,  Inc.,  and considered a trade secret as defined
// in section 499C of the penal code of the State of California.
// Use of  this information  by  anyone  other  than  authorized
// employees of Autodesk, Inc.  is granted  only under a written
// non-disclosure agreement,  expressly  prescribing  the  scope
// and manner of such use.
//
// CREATED BY: scott heath, November, 1993
//
// DESCRIPTION:
//
// This file contains the interface of the AcDbHandle class.
// Instances of this class encapsulate an 8-byte AutoCAD database
// handle value.

#include "adesk.h"

class AcDbHandle
{
public:

    // Misc. exported functions:

    AcDbHandle(void);
    AcDbHandle(const char*);

    AcDbHandle& operator=(const AcDbHandle&);

    AcDbHandle& operator=(const char*);
    void getIntoAsciiBuffer(char* pBuf) const;

    Adesk::Boolean operator == (const AcDbHandle&) const;
    Adesk::Boolean operator != (const AcDbHandle&) const;

    Adesk::Boolean isNull(void) const;

/////// Implementation Class Members: Not for 3rd Party Consumption ///////

    AcDbHandle& operator++(void);
    AcDbHandle operator++(int);

    AcDbHandle operator + (const AcDbHandle&) const;
    AcDbHandle slowOperatorPlus(const AcDbHandle&) const;
    AcDbHandle& operator += (const AcDbHandle&);

    AcDbHandle operator - (const AcDbHandle&) const;
    AcDbHandle& operator -= (const AcDbHandle&);

    Adesk::Boolean operator > (const AcDbHandle&) const;
    Adesk::Boolean operator >= (const AcDbHandle&) const;

    Adesk::Boolean operator < (const AcDbHandle&) const;
    Adesk::Boolean operator <= (const AcDbHandle&) const;

    int compare(const AcDbHandle&) const;

    // Functions to copy handle values between the old and new
    // representations.
    //
    void copyToOldType(Adesk::UInt8 hand[8]) const;
    void copyFromOldType(const Adesk::UInt8 hand[8]);

    // Functions used to convert a handle to/from byte counted
    // buffer format.
    //
    void getValueBytes(Adesk::UInt8*, Adesk::UInt8*) const;
    void setValueBytes(Adesk::UInt8, const Adesk::UInt8*);

    // Conversion to/from Adesk::UInt32.  This is by some special
    // case code to determine offsets between handle values.
    //
    AcDbHandle& operator=(Adesk::UInt32);
    operator Adesk::UInt32() const;
    AcDbHandle operator + (Adesk::UInt32) const;

    void print() const;

    enum {
        kMaxValueBytes = 8
    };

    // Two methods (inline) used for radix sorting during save.
    //
    int byte     (int i) const;
    int restZeros(int i) const;

private:

    // Instance data:

    Adesk::UInt32 mLow;
    Adesk::UInt32 mHigh;

    friend class AcDbHandleTable;
};


extern const AcDbHandle kNullHandle;


// This value is used to indicate cases where a conversion from
// an AcDbHandle to an Adesk::UInt32 has gone bad because the
// handle value was too great.
//
const Adesk::UInt32 kBadUInt32Handle = 0xFFFFFFFF;


inline
AcDbHandle::AcDbHandle(void)
{
}


inline AcDbHandle&
AcDbHandle::operator=(const AcDbHandle& handle)
{
    mHigh = handle.mHigh;
    mLow = handle.mLow;

    return *this;
}


inline AcDbHandle&
AcDbHandle::operator=(Adesk::UInt32 val)
{
    mHigh = 0;
    mLow = val;

    return *this;
}


inline AcDbHandle
AcDbHandle::operator+(const AcDbHandle& handle) const
{
    if ((mHigh | handle.mHigh | ((mLow | handle.mLow) & 0x80000000)) == 0)
    {
	AcDbHandle tHandle;
	tHandle.mHigh = 0;
	tHandle.mLow = mLow + handle.mLow;
	return tHandle;
    }
    else {
	return slowOperatorPlus(handle);
    }
}


inline AcDbHandle
AcDbHandle::operator+(Adesk::UInt32 val) const
{
    if ((mHigh | ((mLow | val) & 0x80000000)) == 0) {
	AcDbHandle tHandle;
	tHandle.mHigh = 0;
	tHandle.mLow = mLow + val;
	return tHandle;
    }
    else {
	AcDbHandle tHandle;
	return slowOperatorPlus(tHandle = val);
    }
}


inline Adesk::Boolean
AcDbHandle::operator > (const AcDbHandle& handle) const
{
    if (mHigh > handle.mHigh)
	return 1;
    else if (mHigh < handle.mHigh)
	return 0;
    else
	return (mLow > handle.mLow);
}


inline int
AcDbHandle::compare(const AcDbHandle& handle) const
{
    if (mHigh > handle.mHigh)
	return -1;
    else if (mHigh < handle.mHigh)
	return 1;

    else if (mLow > handle.mLow)
	return -1;
    else if (mLow < handle.mLow)
	return 1;
    else
	return 0;
}


/************************* Exported Protocol *********************************

AcDbHandle()

    Creates a handle with undefined value.

AcDbHandle(const char*)

    Creates a handle with the value initialized to the value
    described by the hex digits contained in the string.  If
    the string value is invalid (null, too long, or containing
    something other than hex digits) the handle is initialized
    to the "broken handle" value.

AcDbHandle& operator=(const char*)

    Interprets the input string as a set of hex digits indicating
    the desired value for the handle.

int operator == (const AcDbHandle&);
int operator != (const AcDbHandle&);

    Standard behavior.

Adesk::Boolean isNull(void) const

    Returns Adesk::kTrue if the handle is equivalent to the null handle.

*************************** Public Protocol **********************************

AcDbHandle& operator++(void);
AcDbHandle& operator++(int);
AcDbHandle operator + (const AcDbHandle&) const;
AcDbHandle& operator += (const AcDbHandle&);
AcDbHandle operator - (const AcDbHandle&) const;
AcDbHandle& operator -= (const AcDbHandle&);

    Standard behaviour except that underflow or
    overflow results in an internal error.

int operator < (const AcDbHandle&) const;
int operator <= (const AcDbHandle&) const;
int operator > (const AcDbHandle&) const;
int operator >= (const AcDbHandle&) const;

    Standard behavior.

int compare(const AcDbHandle&) const;

    compares two handles and returns -1 if the handle is
    greater than the input handle, 0 of the two handles
    are identical, or 1 if the handle is less than the
    input handle.

void copyToOldType(Adesk::UInt8 output[8]) const;
void copyFromOldType(const Adesk::UInt8 input[8]);

    Functions used to copy values to/from the old format.

AcDbHandle& operator=(const Adesk::UInt32);

    Assigns a handle a value based upon an Adesk::UInt32.

operator Adesk::UInt32() const;

    Converts a handle to an Adesk::UInt32.  If the value
    does not fit it is set to kBadUint32Handle (0xFFFFFFFF).
    As such, the highest meaningfull result is 0xFFFFFFFE.

******************************** Data ****************************************

Adesk::UInt32 mLow;
Adesk::UInt32 mHigh;

    These members contain the low and high 32 bits of the handle value.

******************************************************************************/


// ***** INLINES *****

inline int AcDbHandle::byte(int i) const
{
    if (i < 4)
	return ((mLow >> (i << 3)) & 0xFF);
    else
	return ((mHigh >> ((i - 4) << 3) & 0xFF));
}

inline int AcDbHandle::restZeros(int i) const
{
    Adesk::UInt32 mask;
    if (i < 4) {
	mask = ~0 << (i << 3);
	return !(mHigh | (mLow & mask));
    }
    else {
	mask = ~0 << ((i - 4) << 3);
	return !(mHigh & mask);
    }
}

#endif /* AD_DBHANDLE_H */
