//
//  $Id: dbmtext.h,v 1.24 1995/10/26 16:14:59 neb Exp $
//
//  dbmtext.h - multi-line text entity
//
//  (C) Copyright 1994-1995 by Autodesk, Inc.  All rights reserved.
//
//  The information contained herein is confidential, proprietary to 
//  Autodesk, Inc., and considered a trade secret as defined in 
//  section 499C of the penal code of the State of California.  
//  Use of this information by anyone other than authorized employees 
//  of Autodesk, Inc. is granted only under a written non-disclosure 
//  agreement, expressly prescribing the scope and manner of such use.
//

#ifndef AD_DBMTEXT_H
#define AD_DBMTEXT_H

#include "dbmain.h"

//  text fragment
//  This structure is used to describe a fragment of text as 
//  extracted by the explodeFragments() member function.  Each 
//  fragment consists of a piece of text with a unique 
//  combination of style information (font, size, color, etc.)
struct AcDbMTextFragment
{
    //  position data
    AcGePoint3d location;
    AcGeVector3d normal;
    AcGeVector3d direction;
    //  text characters
    char *text;
    //  font names
    char *font;
    char *bigfont;
    //  metrics
    AcGePoint2d extents;
    double capsHeight;
    double widthFactor;
    double obliqueAngle;
    double trackingFactor;
    Adesk::UInt16 colorIndex;
    //  stack flags
    unsigned stackTop:1;
    unsigned stackBottom:1;
    //  underline/overline data
    unsigned underlined:1;
    unsigned overlined:1;
    AcGePoint3d underPoints[2];
    AcGePoint3d overPoints[2];
};

//  fragment enumerator callback function prototype
//  A pointer to a function of this type is expected by the 
//  explodeFragments() member function.  The elaboration function 
//  is called once for each fragment.
//  'frag' points to the relevant fragment structure.
//  'param' contains the value of the 'param' argument given to 
//  the explodeFragments() member function.
//  The elaboration function must return 1 to continue the 
//  elaboration operation, or 0 to terminate the operation.
typedef int (*AcDbMTextEnum)(AcDbMTextFragment *, void *);

//  AcDbMText class
class AcDbMText: public  AcDbEntity
{
public:
    ACDB_DECLARE_MEMBERS(AcDbMText);
    //  ctor & dtor
    AcDbMText();
    ~AcDbMText();
    //  Insertion point
    AcGePoint3d		location() const;
    void		setLocation(const AcGePoint3d&);
    //  Normal vector
    //  Note that setting the normal vector will automatically 
    //  set a default direction vector.  If you need to set a 
    //  different direction vector or rotation, make those 
    //  calls *after* setting the normal.
    AcGeVector3d normal() const;
    void setNormal(const AcGeVector3d&);
    //  Direction vector
    AcGeVector3d	direction() const;
    void		setDirection(const AcGeVector3d&);
    //  Rotation angle (radians)
    //  The direction vector above is stored with entities.
    //  The following two functions are provided to simplify 
    //  access.  Note that the direction vector need not be 
    //  orthogonal to the normal vector, although calling 
    //  'setRotation()' will create a direction vector that 
    //  is orthogonal.  If the direction vector is non-orthogonal, 
    //  'rotation()' will fail (returning 0.0).
    double		rotation() const;
    void		setRotation(double);
    //  Entity width (constraining value)
    //  If 0.0, disables word wrap
    double		width() const;
    void		setWidth(double);
    //  Text Style (sets initial font)
    AcDbObjectId        textStyle() const;
    void		setTextStyle(AcDbObjectId);
    //  Initial text height (caps height)
    double		textHeight() const;
    void		setTextHeight(double);
    //  
    enum AttachmentPoint
    {
	kTopLeft	= 1,
	kTopCenter	= 2,
	kTopRight	= 3,
	kMiddleLeft	= 4,
	kMiddleCenter	= 5,
	kMiddleRight	= 6,
	kBottomLeft	= 7,
	kBottomCenter	= 8,
	kBottomRight	= 9
    };
    AttachmentPoint	attachment() const;
    void		setAttachment(AttachmentPoint);
    //  
    enum FlowDirection
    {
	kLtoR		= 1,
	kRtoL		= 2,
	kTtoB		= 3,
	kBtoT		= 4
    };
    FlowDirection	flowDirection() const;
    void		setFlowDirection(FlowDirection);
    //  Text contents
    char		*contents() const;
    int			setContents(const char *);
    //  return actual extents
    double		actualHeight() const;
    double		actualWidth() const;
    //  Invoke the selected editor on the entity contents
    int			edit();
    int			correctSpelling();
    //  Return points marking text box
    void getBoundingPoints(AcGePoint3dArray&) const;
    //  Explode text fragments
    void		explodeFragments(AcDbMTextEnum, void *) const;

//  data
    //  These strings can be used instead of sprinkling your code 
    //  with string constants.
    static const char * const nonBreakSpace();	//  \~
    static const char * const overlineOn();	//  \O
    static const char * const overlineOff();	//  \o
    static const char * const underlineOn();	//  \L
    static const char * const underlineOff();	//  \l
    static const char * const colorChange();	//  \C
    static const char * const fontChange();	//  \F
    static const char * const heightChange();	//  \H
    static const char * const widthChange();	//  \W
    static const char * const obliqueChange();	//  \Q
    static const char * const trackChange();	//  \T
    static const char * const lineBreak();	//  \p
    static const char * const paragraphBreak();	//  \P
    static const char * const stackStart();	//  \S
    static const char * const alignChange();	//  \A
    static const char * const blockBegin();	//  {
    static const char * const blockEnd();	//  }

};
MAKE_ACDBOPENOBJECT_FUNCTION(AcDbMText);

#endif
