#ifndef AC_GECURV2D_H
#define AC_GECURV2D_H
//
// $Id: gecurv2d.h,v 1.5 1995/10/26 16:33:18 neb Exp $
//
// (C) Copyright 1993,1995 by Autodesk, Inc. 
//
// The information contained herein is confidential, proprietary  to 
// Autodesk,  Inc., and considered a trade secret as defined in sec- 
// tion 499C of the penal code of the State of California.   Use  of 
// this information by anyone other than authorized employees of Au- 
// todesk, Inc. is  granted  only  under  a  written  non-disclosure 
// agreement,  expressly  prescribing  the  scope and manner of such 
// use.
//
// DESCRIPTION:
//
// This file contains the class AcGeCurve2d - An abstract base class 
// from which all 2d curves are derived.
     
#include "adesk.h"
#include "geent2d.h"
#include "geponc2d.h"
     
class AcGePoint2d;
class AcGeVector2d;
class AcGePointOnCurve2d;
class AcGeInterval;
class AcGeMatrix2d;
class AcGeLine2d;
class AcGePointOnCurve2dData;
class AcGeVector2dArray;
     
class AcGeCurve2d : public AcGeEntity2d
{   
public:
     
    // Parametrization.
    //
    void           getInterval    (AcGeInterval&) const;
    void           getInterval    (AcGeInterval&, AcGePoint2d& start,
				   AcGePoint2d& end) const;
    AcGeCurve2d&   reverseParam   ();
     
    // Distance to other geometric objects. 
    //
    double         distanceTo     (const AcGePoint2d&,
                                   const AcGeTol& = AcGeContext::gTol) const;
    double         distanceTo     (const AcGeCurve2d&,
                                   const AcGeTol& = AcGeContext::gTol) const;

    // Return the point on this object that is closest to the other object.
    // These methods return point on this curve as a simple 2d point.
    //
    AcGePoint2d closestPointTo(const AcGePoint2d& pnt, 
                                  const AcGeTol& = AcGeContext::gTol) const;
    AcGePoint2d closestPointTo(const AcGeCurve2d& curve2d, 
                                  AcGePoint2d& pntOnOtherCrv, 
                                  const AcGeTol& = AcGeContext::gTol) const;


    // Alternate signatures for above functions.  These methods return point
    // on this curve as an AcGePointOnCurve2d.
    //
    void getClosestPointTo(const AcGePoint2d& pnt, 
                                 AcGePointOnCurve2d&, 
                           const AcGeTol& = AcGeContext::gTol) const;
    void getClosestPointTo(const AcGeCurve2d& curve2d, 
                                 AcGePointOnCurve2d& pntOnThisCrv, 
                                 AcGePointOnCurve2d& pntOnOtherCrv, 
                           const AcGeTol& = AcGeContext::gTol) const;
     
    // Tests if point is on curve.
    //
    Adesk::Boolean isOn           (const AcGePoint2d&,
                                   const AcGeTol& = AcGeContext::gTol) const; 
    Adesk::Boolean isOn           (const AcGePoint2d& pt, double& param,
                                   const AcGeTol& = AcGeContext::gTol) const;
    Adesk::Boolean isOn           (double param,
                                   const AcGeTol& = AcGeContext::gTol) const;
     
    // Parameter of the point on curve.  Contract: point IS on curve
    //
    double         paramOf        (const AcGePoint2d& pt,
                                   const AcGeTol& = AcGeContext::gTol) const; 
     
    // Geometric inquiry methods.
    //
    Adesk::Boolean isClosed       (const AcGeTol& = AcGeContext::gTol) const;
    Adesk::Boolean isPeriodic     (double& period) const;
    Adesk::Boolean isLinear       (AcGeLine2d&,
                                   const AcGeTol& = AcGeContext::gTol) const;

    // Length based methods.
    //
    double         length         (double fromParam, double toParam,
                                   double = AcGeContext::gTol.equalPoint())
                                  const;
    double         paramAtLength  (double datumParam, double length,
                                   Adesk::Boolean posParamDir = Adesk::kTrue,
                                   double = AcGeContext::gTol.equalPoint())
                                  const;     
    // Degeneracy.
    //
    Adesk::Boolean isDegenerate   (AcGe::EntityId& degenerateType,
                                   const AcGeTol& = AcGeContext::gTol)
                                  const;
    Adesk::Boolean isDegenerate   (AcGeEntity2d*& pConvertedEntity,
                                   const AcGeTol& = AcGeContext::gTol) const;

    // Modify methods.
    //                                                  
    void           getSplitCurves (double param, AcGeCurve2d* & piece1, 
                                   AcGeCurve2d* & piece2) const;
     
    // Local closest points
    //
    void getLocalClosestPoints(const AcGePoint2d& point, 
                                     AcGePointOnCurve2d& approxPnt, 
                               const AcGeInterval* nbhd = 0,
                               const AcGeTol& = AcGeContext::gTol) const;
     
    void getLocalClosestPoints(const AcGeCurve2d& otherCurve, 
                                     AcGePointOnCurve2d& approxPntOnThisCrv, 
                                     AcGePointOnCurve2d& approxPntOnOtherCrv, 
                               const AcGeInterval* nbhd1 = 0,
                               const AcGeInterval* nbhd2 = 0,
                               const AcGeTol& = AcGeContext::gTol) const;
    
    // Return start and end points.
    //
    Adesk::Boolean hasStartPoint  (AcGePoint2d& startPoint) const;
    Adesk::Boolean hasEndPoint    (AcGePoint2d& endPoint) const;
    
    // Evaluate methods.
    //
    AcGePoint2d    evalPoint   (double param) const;
    AcGePoint2d    evalPoint   (double param, int numDeriv,
                                AcGeVector2dArray& derivArray) const;

    // Assignment operator.
    //
    AcGeCurve2d&   operator =  (const AcGeCurve2d&);

protected:

    friend class AcGeImpCurve2d;

    // Private constructors so that no object of this class can be instantiated.
    AcGeCurve2d ();
    AcGeCurve2d (const AcGeCurve2d&);

    // Construct object from its corresponding implementation object.
    AcGeCurve2d (AcGeImpCurve2d&, int);
    AcGeCurve2d (AcGeImpCurve2d*);
};

#endif
