#ifndef AC_GEINTRVL_H
#define AC_GEINTRVL_H
//
// $Id: geintrvl.h,v 1.6 1995/09/21 11:21:34 jamesr Exp $
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary  to
// Autodesk,  Inc., and considered a trade secret as defined in sec-
// tion 499C of the penal code of the State of California.   Use  of
// this information by anyone other than authorized employees of Au-
// todesk, Inc. is  granted  only  under  a  written  non-disclosure
// agreement,  expressly  prescribing  the  scope and manner of such
// use.
//
// DESCRIPTION:
//
// This file contains the class AcGeInterval - a representation
// for an interval on the real line.  The following kinds of intervals
// are supported.
// . Open intervals where the bounds not finite.
// . Closed intervals with finite bounds.
// . Half intervals such that one end is open and not finite,
//   and the other is closed and finite.
// 
// Where applicable, all evaluations are performed within the
// tolerance stored within this class.
//

#include "gegbl.h"


class AcGeInterval
{
public:
    AcGeInterval(double tol = 1.e-12);
    AcGeInterval(const AcGeInterval&);
    AcGeInterval(double lower, double upper, double tol = 1.e-12);
    AcGeInterval(Adesk::Boolean boundedBelow, double bound,
                 double tol = 1.e-12);

    // Destructor.
    //
    ~AcGeInterval();

    // Assignment operator.
    //
    AcGeInterval&  operator =       (const AcGeInterval&);
    
    // Get/set methods.
    //
    double         lowerBound       () const;
    double         upperBound       () const;
    double         element          () const;
    void           getBounds        (double& lower, double& upper) const;
    double         length           () const;
    double         tolerance        () const;
    
    AcGeInterval&  set              (double lower, double upper);
    AcGeInterval&  set              (Adesk::Boolean boundedBelow, double bound);
    AcGeInterval&  set              ();
    AcGeInterval&  setUpper         (double);
    AcGeInterval&  setLower         (double);
    AcGeInterval&  setTolerance     (double);                                     
    
    // Interval editing.
    //
    void           getMerge         (const AcGeInterval&, AcGeInterval&) const;
    int            subtract         (const AcGeInterval&, 
                                     AcGeInterval& lInterval,
                                     AcGeInterval& rInterval) const;
    Adesk::Boolean intersectWith    (const AcGeInterval&, AcGeInterval&) const;

    // Interval characterization.
    //
    Adesk::Boolean isBounded        () const;
    Adesk::Boolean isBoundedAbove   () const;
    Adesk::Boolean isBoundedBelow   () const;
    Adesk::Boolean isUnBounded      () const;
    Adesk::Boolean isSingleton      () const;

    // Relation to other intervals.
    //
    Adesk::Boolean isDisjoint       (const AcGeInterval&) const;
    Adesk::Boolean contains         (const AcGeInterval&) const;
    Adesk::Boolean contains         (double) const;

    // Continuity
    //
    Adesk::Boolean isContinuousAtUpper (const AcGeInterval&) const;
    Adesk::Boolean isOverlapAtUpper    (const AcGeInterval&,
                                        AcGeInterval& overlap) const;
    // Equality
    //
    Adesk::Boolean operator ==      (const AcGeInterval&) const;
    Adesk::Boolean operator !=      (const AcGeInterval&) const;
    Adesk::Boolean isEqualAtUpper   (const AcGeInterval&) const;
    Adesk::Boolean isEqualAtUpper   (double value) const;
    Adesk::Boolean isEqualAtLower   (const AcGeInterval&) const;
    Adesk::Boolean isEqualAtLower   (double value) const;

    // To be used with periodic curves
    //
    Adesk::Boolean isPeriodicallyOn (double period, double& t);

    // Comparisons.
    //
    friend
    Adesk::Boolean operator >       (double, const AcGeInterval&);
    Adesk::Boolean operator >       (double) const;
    Adesk::Boolean operator >       (const AcGeInterval&) const;
    friend
    Adesk::Boolean operator >=      (double, const AcGeInterval&);
    Adesk::Boolean operator >=      (double) const;
    Adesk::Boolean operator >=      (const AcGeInterval&) const;
    friend
    Adesk::Boolean operator <       (double, const AcGeInterval&);
    Adesk::Boolean operator <       (double) const;
    Adesk::Boolean operator <       (const AcGeInterval&) const;
    friend
    Adesk::Boolean operator <=      (double, const AcGeInterval&);
    Adesk::Boolean operator <=      (double) const;
    Adesk::Boolean operator <=      (const AcGeInterval&) const;

protected:
    friend class AcGeImpInterval;

    AcGeImpInterval  *mpImpInt;

    // Construct object from its corresponding implementation object.
    AcGeInterval (AcGeImpInterval&, int);
    AcGeInterval (AcGeImpInterval*);
                                         
private:
    int              mDelInt;
};


#endif
