#ifndef AC_GEKVEC_H
#define AC_GEKVEC_H
//
// $Id: gekvec.h,v 1.3 1995/10/26 16:34:44 neb Exp $
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary  to
// Autodesk,  Inc., and considered a trade secret as defined in sec-
// tion 499C of the penal code of the State of California.   Use  of
// this information by anyone other than authorized employees of Au-
// todesk, Inc. is  granted  only  under  a  written  non-disclosure
// agreement,  expressly  prescribing  the  scope and manner of such
// use.
//
//
// DESCRIPTION:
//
// This file contains the class AcGeKnotVector - a dynamic array used
// to represent the knot sequence of a spline.  Refer to
// AcGeDoubleArray for details concerning the dynamic array
// behavior.

#include "gegbl.h"
#include "gedblar.h"

class AcGeInterval;

class AcGeKnotVector
{
public:
    AcGeKnotVector(double eps = 1.e-9);
    AcGeKnotVector(int size, int growSize, double eps = 1.e-9);
	AcGeKnotVector(int size, const double [], double eps = 1.e-9);
    AcGeKnotVector(const AcGeKnotVector& src);
    AcGeKnotVector(const AcGeDoubleArray& src, double eps = 1.e-9);

    ~AcGeKnotVector();

    // Copy operator.
    //
    AcGeKnotVector&     operator =  (const AcGeKnotVector&);
	AcGeKnotVector&     operator =  (const AcGeDoubleArray&);


    // Indexing into the knot vector.
    //
    double&             operator [] (int);
    const double        operator [] (int) const;
    
    // Equality test
    //
    Adesk::Boolean      isEqualTo (const AcGeKnotVector& other) const;

    // Inquiry functions
    //
    double              startParam         () const;
    double              endParam           () const;
    int                 multiplicityAt     (int i) const;
    int                 numIntervals       () const;

    // Evaluate funtions
    //
    int                 getInterval        (int ord, double par, 
                                            AcGeInterval& interval ) const;
    void                getDistinctKnots   (AcGeDoubleArray&) const;
    Adesk::Boolean      contains           (double) const;

    // Edit function
    //
    AcGeKnotVector&     reverse            ();
    AcGeKnotVector&     removeAt           (int);
    AcGeKnotVector&     insertAt           (int indx, double u,
                                            int multiplicity = 1);
    AcGeKnotVector&     insert             (double);
    int                 append             (double); 
    AcGeKnotVector&     append             (AcGeKnotVector& tail, 
                                            double knotRatio = 0.);
    int                 split              (double par,
                                            AcGeKnotVector* pKnot1, 
                                            int multLast,
                                            AcGeKnotVector* pKnot2, 
                                            int multFirst ) const;


    AcGeKnotVector&     setRange           (double lower, double upper);

	double              tolerance          () const;
	AcGeKnotVector&     setTolerance       (double);

	// Array length.
	//
	int                 length             () const; // Logical length.
    Adesk::Boolean      isEmpty            () const;
    int                 logicalLength      () const;
    AcGeKnotVector&     setLogicalLength   (int);
    int                 physicalLength     () const;
    AcGeKnotVector&     setPhysicalLength  (int);

    // Automatic resizing.
    //
    int                 growLength  () const;
    AcGeKnotVector&     setGrowLength(int);

	// Treat as simple array of double.
    //
    const double*       asArrayPtr  () const;
    double*             asArrayPtr  ();

	AcGeKnotVector&     set (int size, const double [], double eps = 1.e-9);

protected:
    AcGeDoubleArray    mData;
    double             mTolerance;

    Adesk::Boolean     isValid (int) const;
};

// Inline methods.
//
inline double
AcGeKnotVector::tolerance() const
{ return mTolerance;}

inline AcGeKnotVector&
AcGeKnotVector::setTolerance(double eps)
{ mTolerance = eps;	return *this;}

inline Adesk::Boolean
AcGeKnotVector::isValid(int i) const
{ return i >= 0 && i < mData.logicalLength(); }

inline double&
AcGeKnotVector::operator [] (int i)
{ assert(isValid(i)); return mData[i]; }

inline const double
AcGeKnotVector::operator [] (int i) const
{ assert(isValid(i)); return mData[i]; }

inline const double*
AcGeKnotVector::asArrayPtr() const
{ return mData.asArrayPtr(); }

inline double*
AcGeKnotVector::asArrayPtr()
{ return mData.asArrayPtr(); }



#endif
