#ifndef AC_GEMAT2D_H
#define AC_GEMAT2D_H
//
// $Id: gemat2d.h,v 1.1 1995/07/28 16:53:31 tomba Exp $
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary  to
// Autodesk,  Inc., and considered a trade secret as defined in sec-
// tion 499C of the penal code of the State of California.   Use  of
// this information by anyone other than authorized employees of Au-
// todesk, Inc. is  granted  only  under  a  written  non-disclosure
// agreement,  expressly  prescribing  the  scope and manner of such
// use.
//
// DESCRIPTION:
//
// This file contains the class AcGeMatrix2d - a matrix used to
// represent general purpose 2d transformations in homogeneous
// co-ordinate space.

#include "gegbl.h"
#include "gepnt2d.h"

class AcGeVector2d;
class AcGeLine2d;
class AcGeTol;

class AcGeMatrix2d
{
public:
    AcGeMatrix2d();
    AcGeMatrix2d(const AcGeMatrix2d&);

    // The multiplicative identity.
    //
    static const   AcGeMatrix2d   kIdentity;

    // Reset matrix.
    //
    AcGeMatrix2d&  setToIdentity();

    // Multiplication.
    //
    AcGeMatrix2d   operator *   (const AcGeMatrix2d&) const;
    AcGeMatrix2d&  operator *=  (const AcGeMatrix2d&);
    AcGeMatrix2d&  preMultBy    (const AcGeMatrix2d& leftSide);
    AcGeMatrix2d&  postMultBy   (const AcGeMatrix2d& rightSide);
    AcGeMatrix2d&  setToProduct (const AcGeMatrix2d&, const AcGeMatrix2d&);

    // Multiplicative inverse.
    //
    AcGeMatrix2d&  invert       ();
    AcGeMatrix2d   inverse      () const;

    // Test if it is a singular matrix.  A singular matrix is not invertable.
    // 
    Adesk::Boolean isSingular   (const AcGeTol& = AcGeContext::gTol) const;

    // Matrix transposition.
    //
    AcGeMatrix2d&  transposeIt  ();
    AcGeMatrix2d   transpose    () const;

    // Tests for equivalence using the infinity norm.
    //
    Adesk::Boolean operator ==  (const AcGeMatrix2d&) const;
    Adesk::Boolean operator !=  (const AcGeMatrix2d&) const;
    Adesk::Boolean isEqualTo    (const AcGeMatrix2d&, const AcGeTol& = 
                                       AcGeContext::gTol) const;

    // Test scaling effects of matrix
    //
    Adesk::Boolean isUniScaledOrtho(const AcGeTol& = AcGeContext::gTol) const;
    Adesk::Boolean isScaledOrtho   (const AcGeTol& = AcGeContext::gTol) const;

    // Determinant
    //
    double         det          () const;

    // Set/retrieve translation.
    //
    AcGeMatrix2d&  setTranslation(const AcGeVector2d&);
    AcGeVector2d   translation   () const;

    // Set/get coordinate system
    //
    AcGeMatrix2d&  setCoordSystem(const AcGePoint2d& origin,
                                  const AcGeVector2d& e0,
                                  const AcGeVector2d& e1);
    void           getCoordSystem(AcGePoint2d& origin, 
                                  AcGeVector2d& e0,
                                  AcGeVector2d& e1) const;

    // Set the matrix to be a specified transformation 
    //
    AcGeMatrix2d& setToTranslation   (const AcGeVector2d&);
    AcGeMatrix2d& setToRotation      (double angle, 
                                      const AcGePoint2d& center = 
                                             AcGePoint2d::kOrigin);
    AcGeMatrix2d& setToScaling       (double scaleAll, 
                                      const AcGePoint2d& center = 
                                             AcGePoint2d::kOrigin);
    AcGeMatrix2d& setToMirroring     (const AcGePoint2d&);
    AcGeMatrix2d& setToMirroring     (const AcGeLine2d&);
    AcGeMatrix2d& setToAlignCoordSys (const AcGePoint2d&  fromOrigin,
                                      const AcGeVector2d& fromE0,
                                      const AcGeVector2d& fromE1,
                                      const AcGePoint2d&  toOrigin,
                                      const AcGeVector2d& toE0,
                                      const AcGeVector2d& toE1);

    // Functions that make a 2d transformation matrix using various approaches
    //
    static 
    AcGeMatrix2d translation   (const AcGeVector2d&);
    static 
    AcGeMatrix2d rotation      (double angle, 
                                       const AcGePoint2d& center = 
                                              AcGePoint2d::kOrigin);
    static 
    AcGeMatrix2d scaling       (double scaleAll, 
                                       const AcGePoint2d& center = 
                                              AcGePoint2d::kOrigin);
    static 
    AcGeMatrix2d mirroring     (const AcGePoint2d&);
    static 
    AcGeMatrix2d mirroring     (const AcGeLine2d&);
    static 
    AcGeMatrix2d alignCoordSys  (const AcGePoint2d&  fromOrigin,
                                       const AcGeVector2d& fromE0,
                                       const AcGeVector2d& fromE1,
                                       const AcGePoint2d&  toOrigin,
                                       const AcGeVector2d& toE0,
                                       const AcGeVector2d& toE1);
    // For convenient access to the data.
    //
    double         operator     () (unsigned int, unsigned int) const;
    double&        operator     () (unsigned int, unsigned int);

    // The components of the matrix.
    //
    double         entry[3][3]; // [row][column]
};

// Return a reference to the element in position [row][column]
// of the `entry' array.
//
inline double
AcGeMatrix2d::operator () (
    unsigned int row,
    unsigned int column) const
{
    return entry[row][column];
}

inline double&
AcGeMatrix2d::operator () (
    unsigned int row,
    unsigned int column)
{
    return entry[row][column];
}

#endif
