// $Id: design.cc,v 1.7 1995/07/14 14:22:31 elevy Exp $
//*****************************************************************************
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary
// to Autodesk,  Inc.,  and considered a trade secret as defined
// in section 499C of the penal code of the State of California.
// Use of  this information  by  anyone  other  than  authorized
// employees of Autodesk, Inc.  is granted  only under a written
// non-disclosure agreement,  expressly  prescribing  the  scope
// and manner of such use.
//
// CREATED BY:  Anand M. Das                  April 24, 1994
//
// DESCRIPTION: Source file for the Rx application "SHAFT"
//
//*****************************************************************************
 
#include "sample.h"
#include "aced.h"
#include "dbents.h"
#include "adslib.h"
#include "acdb.h"
#include "dbsol3d.h"
#include "gemat3d.h"


extern "C" {
    // locally defined entry point invoked by Rx.
    AcRx::AppRetCode acrxEntryPoint ( AcRx::AppMsgCode msg, void* );
}

void initModule();

/* =================== Rx Module Interface Functions ================ */

AcRxDictionary* acdbCmds = NULL;

AcRx::AppRetCode acrxEntryPoint ( AcRx::AppMsgCode msg, void* ) {
    switch ( msg ) {
    case AcRx::kInitAppMsg:
	initModule();
	break;
    case AcRx::kUnloadAppMsg:
        acedRegCmds->removeGroup ( "SHAFT" );
	delete acdbCmds;
        break;
    default:
	break;
    }
    return AcRx::kRetOK;
}


// Initialization Function

void initModule()
{
    // Add some commands to test the AcDb API Library

    acedRegCmds->addCommand("SHAFT", "SHAFT", "SHAFT", ACRX_CMD_MODAL,
        &createShaft);
}



//***********************************************************************
//***********************************************************************
//
//        This function executes the SHAFT command
//
//***********************************************************************
//***********************************************************************


void createShaft()
{
    double                           shaftDiameter, shaftLength, keyDepth,
                                     keyWidth, keyLength, shaftRadius;
    AcGePoint3d                      shaftPoint;
    AcDb3dSolid*                     shaftObj = NULL;
    AcDb3dSolid*                     blockObj = NULL;
    AcGePoint3d                      shaftCenter;
    AcGePoint3d                      blockCenter;
    AcGeMatrix3d                     shaftXform, transMat, xformMat;
    AcDbObjectId                     shaftId;

    // Prompt user for shaft data

    if ( getShaftData ( shaftDiameter, shaftLength, keyDepth, keyWidth,
 	keyLength, shaftPoint ) != Acad::eOk )
        goto Error;

    // Create the cylindrical shaft with the centroid at the 
    // world coordinate system

    shaftRadius = shaftDiameter * 0.5;

    if (( shaftObj = new AcDb3dSolid ) == NULL )
        goto Error;

    if ( shaftObj->createFrustum ( shaftLength, shaftRadius,
        shaftRadius, shaftRadius ) != Acad::eOk )
        goto Error;

    // Create a block of keyway size 

    if (( blockObj = new AcDb3dSolid ) == NULL )
        goto Error;

    if ( blockObj->createBox ( keyDepth, keyWidth, keyLength ) != Acad::eOk )
        goto Error;


    // Following operations are needed to transform the block and the 
    // cylindrical shaft to the correct locations.

    // Get the coordinates of the points to where the center of the
    // shaft and the center of the block have to be moved.

    shaftCenter.set ( shaftPoint.x, shaftPoint.y, 
        ( shaftPoint.z + ( 0.5 * shaftLength ) ) );

    blockCenter.set ( shaftPoint.x + ( (shaftDiameter - keyDepth) * 0.5 ),
        shaftPoint.y, ( shaftPoint.z + ( keyLength * 0.5 ) ) );

    // Get the UCS transformation matrix

    if ( getWorldToUserTransform ( shaftXform ) != Acad::eOk )
        goto Error;

    // Get the transformation matrix to be applied to the shaft

    transMat.entry[0][3] = shaftCenter.x;
    transMat.entry[1][3] = shaftCenter.y;
    transMat.entry[2][3] = shaftCenter.z;

    xformMat = shaftXform * transMat;

    if ( shaftObj->transformBy ( xformMat ) != Acad::eOk )
        goto Error;

    // Get the transformation matrix to be applied to the block

    transMat.entry[0][3] = blockCenter.x;
    transMat.entry[1][3] = blockCenter.y;
    transMat.entry[2][3] = blockCenter.z;

    xformMat = shaftXform * transMat;

    if ( blockObj->transformBy ( xformMat ) != Acad::eOk )
        goto Error;

    // Now, subtract the block from the cylinder to create the keyway

    if ( shaftObj->booleanOper ( AcDb::kBoolSubtract, blockObj ) 
        != Acad::eOk )
        goto Error;

    if ( postObject ( shaftObj, shaftId ) != Acad::eOk )
        goto Error;

    // All went well

    return;

Error:

    ads_printf ( "\n Shaft creation failed \n" );

    // delete all created objects

    delete shaftObj;
    delete blockObj;
 
    return;
}


//***********************************************************************
//***********************************************************************
//
//   This routine prompts the user to enter shaft data. Default values
//   are supplied to the user.
//
//***********************************************************************
//***********************************************************************



Acad::ErrorStatus 
getShaftData ( double& dia, double& len, double& keyd,
               double& keyw, double& keyl, AcGePoint3d& shaftPoint )
{
    double                         temp, keywMax;
    ads_point                      adsPoint;
    int                            stat; // ADS error status 

    // get the point where the user wants to construct the shaft

    shaftPoint.set ( 0.0, 0.0, 0.0 );
    if ( ( stat = ads_getpoint (NULL, "\nInsertion base point <0,0,0>: ", 
        adsPoint ) ) != RTNORM )
        if ( stat != RTNONE )
            goto Error;
			 
    // convert ads_point into AcGePoint3d

    shaftPoint = asPnt3d ( adsPoint );

    // get all the other shaft data

    dia = 1.0;
    ads_initget ( 2+4, NULL ); // allow only values greater than zero
    if ( ( stat = ads_getreal ( "\nEnter shaft diameter <1.0>: ", &dia ) )
        !=  RTNORM  )
        if ( stat != RTNONE )
            goto Error;

    len = 5.0;
    ads_initget ( 2+4, NULL ); // allow only values greater than zero    
    if ( ( stat = ads_getreal ( "\nEnter shaft length <5.0>: ", &len ) )
        !=  RTNORM  )
        if ( stat != RTNONE )
            goto Error;

    do {
        keyd = 0.3;
        ads_initget ( 2+4, NULL ); // allow only values greater than zero    
        if ( ( stat = ads_getreal ( "\nEnter keyway depth <0.3>: ", &keyd ) )
            !=  RTNORM  )
            if ( stat != RTNONE )
                goto Error;
        if ( keyd >= dia )
            ads_printf( "\n Value must be less than the shaft diameter - %f", dia );
    } while ( keyd >= dia );

    temp = ( dia * keyd ) - ( keyd * keyd );

    keywMax = 2.0 * sqrt( temp );

    do {
        keyw = 0.4;
        ads_initget ( 2+4, NULL ); // allow only values greater than zero    
        if ( ( stat =  ads_getreal ( "\nEnter keyway width <0.4>: ", &keyw ) )
            != RTNORM )
            if ( stat != RTNONE )
                goto Error;
        if ( keyw >= keywMax )
            ads_printf( "\nValue must be less than %f.", keywMax );
    } while ( keyw >= keywMax ) ;

    do {
        keyl = 2.0;
        ads_initget ( 2+4, NULL ); // allow only values greater than zero    
        if ( ( stat = ads_getreal ( "\nEnter keyway length <2.0>: ", &keyl ) )
            !=  RTNORM )
            if ( stat != RTNONE )       
                goto Error;
        if ( keyl >= len )
            ads_printf( "\nValue must be less than the shaft length - %f", len );
    } while ( keyl >=len );

    return Acad::eOk;

Error:

    ads_printf ( "\n Severe Error in getShaftData() " );

    return Acad::eUnrecoverableErrors;
}






