// $Id: poly.cc,v 1.27 1995/12/20 17:05:05 neb Exp $
//
// (C) Copyright 1993-1995 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary
// to Autodesk,  Inc.,	and considered a trade secret as defined
// in section 499C of the penal code of the State of California.
// Use of  this information  by  anyone  other	than  authorized
// employees of Autodesk, Inc.	is granted  only under a written
// non-disclosure agreement,  expressly  prescribing  the  scope
// and manner of such use.
//
// POLY.CC 
//
// CREATED BY: Keshav Sahoo		     May 1994 
//
// DESCRIPTION:
//
// This file contains the body of the class AcDbPolygon.

#include <stdlib.h>

#include "stdio.h"
#include "string.h"
#include "assert.h"
#include "math.h"

#include "rxobject.h"
#include "rxclass.h"
#include "poly.h"

#include "aced.h"
#include "actrans.h"

#include "acgi.h"

#include "gepnt3d.h"
#include "gevec3d.h"
#include "gemat3d.h"
#include "gelnsg3d.h"

#include "gearc2d.h"
#include "gearc3d.h"
#include "geplane.h"

#include "geassign.h"
#include "gegbl.h"

#include "dbents.h"
#include "dbspline.h"

#include "gegblabb.h"

#include "util.h"
#include "dbsymtb.h"

#include "acdbabb.h"
#include "adeskabb.h"



// extern "C" char* strdup(const char*);

//*************************************************************************
// Statics functions used in this file. 
//*************************************************************************

static Acad::ErrorStatus drawFacets(const AcDbPoly*        poly,
				    AcGiWorldDraw*         worldDraw,
				    AcGiViewportDraw*      vportDraw);

static Acad::ErrorStatus drawEdges (const AcDbPoly*        poly,
				    AcGiWorldDraw*         worldDraw,
				    AcGiViewportDraw*      vportDraw);

static Acad::ErrorStatus intLine   (const AcDbPoly*        poly, 
				    const AcGeLine3d       line,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intLine   (const AcDbPoly*        poly, 
				    const AcDbLine*        line,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intLine   (const AcDbPoly*        poly, 
				    const AcGeLineSeg3d    line,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intArc    (const AcDbPoly*        poly, 
				    const AcDbArc*         arc,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intArc    (const AcDbPoly*        poly, 
				    const AcGeCircArc3d    arc,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intCircle (const AcDbPoly*        poly, 
				    const AcDbCircle*      circle,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intCircle (const AcDbPoly*        poly, 
				    const AcGeCircArc3d     circle,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intPline  (const AcDbPoly*        poly, 
				    AcDb2dPolyline*        pline,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus intPline  (const AcDbPoly*        poly, 
				    AcDb3dPolyline*        pline,
				    AcDb::Intersect        intType,
				    const AcGePlane*       projPlane,
				    AcGePoint3dArray&         points);

static Acad::ErrorStatus drawName(const AcDbPoly*    poly,
				  AcGiWorldDraw*     worldDraw);


//*************************************************************************
// Code for the Class Body. 
//*************************************************************************


ACRX_DXF_DEFINE_MEMBERS(AcDbPoly, AcDbCurve, 0, POLYGON, /*MSG0*/"AutoCAD");


//*************************************************************************
// Constructors and destructors 
//*************************************************************************

AcDbPoly::AcDbPoly()
{
    mCenter.set(0.0, 0.0, 0.0);
    mStartPoint.set(0.0, 0.0, 0.0);
    mNumSides = 0;
    mPlaneNormal.set(0.0, 0.0, 1.0);
    mTextStyle = AcDbObjectId::kNull;
    mpName = NULL;
}

AcDbPoly::AcDbPoly(const AcGePoint3d&   center,
                   const AcGePoint3d&   startPoint,
	           int                  numSides, 
                   const AcGeVector3d&  normal,
		   const char*		name)
{
    mCenter	   = center;
    mStartPoint    = startPoint;
    mNumSides	   = numSides;
    mPlaneNormal   = normal;
    mpName	   = strdup(name);
    mTextStyle	   = AcDbObjectId::kNull;
}

AcDbPoly::AcDbPoly(const AcGePoint3d&   center,
                   const AcGePoint3d&   startPoint,
	           int                  numSides, 
                   const AcGeVector3d&  normal,
		   const char*		name,
		   AcDbObjectId         styleId)
{
    mCenter	   = center;
    mStartPoint    = startPoint;
    mNumSides	   = numSides;
    mPlaneNormal   = normal;
    mpName	   = strdup(name);
    mTextStyle	   = styleId;
}

AcDbPoly::~AcDbPoly()
{
    if (mpName != NULL)
	free(mpName);
}


//*************************************************************************
// Methods specific to AcDbPoly 
//*************************************************************************

char*
AcDbPoly::name() const
{
    assertReadEnabled();
    return mpName == NULL ? NULL : strdup(mpName);
}

AcDbObjectId
AcDbPoly::styleId() const
{
    assertReadEnabled();
    return mTextStyle;
}


Acad::ErrorStatus
AcDbPoly::setName(const char* pName)
{
    assertWriteEnabled();
    if (mpName != NULL)
	free(mpName);
    mpName = strdup(pName);
    return Acad::eOk;
}


Acad::ErrorStatus  
AcDbPoly::setNumSides(int numSides)    
{
    if (mNumSides == numSides)
	return eOk;

    assertWriteEnabled(kFalse, kTrue);
    
    // There are situations under which AutoCAD doesn't want to do undo 
    // recording, so it won't create an undo filer. Check for the existence
    // of the filer before starting to write into it.
    //
    AcDbDwgFiler *pFiler = NULL;
    if ((pFiler = undoFiler()) != NULL) {
        undoFiler()->writeItem((long)AcDbPoly::desc());
        undoFiler()->writeItem((Adesk::Int16)kSetNumSides);
        undoFiler()->writeItem((Adesk::Int32)mNumSides);
    }

    mNumSides = numSides; 
    return Acad::eOk;
}


Acad::ErrorStatus
AcDbPoly::setTextStyle(AcDbObjectId tsId)
{
    assertWriteEnabled();
    mTextStyle = tsId;
    return Acad::eOk;
}


AcGeVector3d	   
AcDbPoly::sideVector	(int  whichSide)	     const
{
    assertReadEnabled();

    AcGePoint3dArray vertexArray;
    Acad::ErrorStatus es = getVertices(vertexArray);

    AcGeVector3d vec = vertexArray[whichSide + 1] - vertexArray[whichSide];
    assert(!vec.isZeroLength());
    vec.normalize();

    return vec;
}


Acad::ErrorStatus 
AcDbPoly::getCoordSys(AcGeMatrix3d& mat) const
{
    assertReadEnabled();

    AcGeVector3d x, y;

    x = mStartPoint - mCenter;
    x.normalize();
    y = mPlaneNormal.crossProduct(x);

    mat.setCoordSystem(mCenter, x, y, mPlaneNormal);

    return Acad::eOk;
}


Acad::ErrorStatus 
AcDbPoly::getVertices(AcGePoint3dArray& vertexArray) const
{
    assertReadEnabled();

    // Calculate the polygon vertices assuming the center of the 
    // polygon to be at the origin and the start point being on the
    // x axis. Then calculate a transformation between the coordinate
    // system that can be formed using the center, start point and 
    // the normal and the world. Apply the transformation to all
    // the points.
    //
    // This may not be the most efficient way of getting the points
    // as far as computation is concerned, but it sure is easy to
    // implement, and it's chosen for that reason.

    double circumRadius = (mStartPoint - mCenter).length();
    double sideLength = 2.0 * circumRadius * cos(3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides);

    AcGePoint3d vertex(circumRadius, 0.0, 0.0);
    double rotateBy, increment;
    
    vertexArray.append(vertex);
    rotateBy = increment = 2.0 * 3.14159265358979323846 / mNumSides;

    for (int i = 0; i < mNumSides - 1; i++) {
	
	vertex[X] = circumRadius * cos(rotateBy); 
	vertex[Y] = circumRadius * sin(rotateBy); 
	vertexArray.append(vertex);
	rotateBy += increment;
    }

    vertex.set(circumRadius, 0.0, 0.0);
    vertexArray.append(vertex);

    // Now calculate the transformation matrix and transform all the points.

    AcGeMatrix3d mat;
    this->getCoordSys(mat);

    for (i = 0; i < vertexArray.length(); i++) {
	vertexArray[i].transformBy(mat);
    }

    return Acad::eOk;
}


//*************************************************************************
// Overridden methods from AcDbCurve
//*************************************************************************


Acad::ErrorStatus 
AcDbPoly::getPointAtParam (double             param, 
                        AcGePoint3d&       point) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Find out how many included angles does "param" contain.
    // Find that vertex, and the direction of the line defined
    // by that vertex and the following vertex. Add the appropriate
    // distance along that line direction to get the new point.

    rx_fixangle(param);

    double circumRadius      = (mStartPoint - mCenter).length();
    double includedAngle     = 6.28318530717958647692 / mNumSides;
    int    numIncludedAngles = param / includedAngle;
    double theta	     = param - numIncludedAngles * includedAngle;

    double gamma	     = 3.14159265358979323846 * 0.5 + 3.14159265358979323846 / mNumSides - theta;
    double distToGo	     = circumRadius * sin(theta) / sin(gamma);

    AcGeVector3d lineDir     = this->sideVector(numIncludedAngles);

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    if (theta > 1.0e-10) {
	point = vertexArray[numIncludedAngles] + (distToGo * lineDir);
    } else {
	point = vertexArray[numIncludedAngles];
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getParamAtPoint (const AcGePoint3d& point,
                           double&            param) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Check to see if the point is the same as one of the vertices of
    // the polygon. If so, great. Otherwise, see if it lies in between 
    // two consecutive vertices. If neither is true, the point if
    // not on the polygon.

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }
   
    double includedAngle = 6.28318530717958647692 / mNumSides;
    double circumRadius  = (mStartPoint - mCenter).length();

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	if (point == vertexArray[i]) {
	    param = i * includedAngle;
	    return Acad::eOk;
	} else if (point == vertexArray[i + 1]) {
	    param = (i + 1) * includedAngle;
	    return Acad::eOk;
	} else {
	    
	    AcGeVector3d v1, v2;
	    v1 = point - vertexArray[i];
	    v2 = vertexArray[i + 1] - point;

	    if (v2.isCodirectionalTo(v1)) {

		double alpha	      = 3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides;
		double halfSideLength = circumRadius * cos(alpha);
		double delta	      = atan(
					fabs(v1.length() - halfSideLength) /
					(circumRadius * sin(alpha)));

		param = v1.length() > halfSideLength ? 
			((i + 0.5) * includedAngle + delta)  :
			((i + 0.5) * includedAngle - delta);

		return Acad::eOk;
	    }
	}
    }

    return Acad::ePointNotOnEntity;
}


Acad::ErrorStatus 
AcDbPoly::getDistAtParam  (double		   param,
		           double& 	    dist)  const
{
    assertReadEnabled();

    rx_fixangle(param);

    double circumRadius = (mStartPoint - mCenter).length();
    double alpha	= 3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides;
    double sideLength	= 2.0 * circumRadius * cos(alpha);

    double includedAngle     = 6.28318530717958647692 / mNumSides;
    int    numIncludedAngles = param / includedAngle;
    double theta	     = param - numIncludedAngles * includedAngle;

    if (theta > 1.0e-10) {
	dist = sideLength * numIncludedAngles +
	       circumRadius * sin(theta) / sin(3.14159265358979323846 - theta - alpha);
    } else {
	dist = sideLength * numIncludedAngles;
    }

    return Acad::eOk;
}


Acad::ErrorStatus 
AcDbPoly::getParamAtDist  (double		   dist,
			   double& 	   param) const 
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    double circumRadius  = (mStartPoint - mCenter).length();
    double alpha	 = 3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides;
    double sideLength	 = 2.0 * circumRadius * cos(alpha);

    int    numSides	 = dist / sideLength;
    double distToGo	 = dist - numSides * sideLength;

    double includedAngle = 6.28318530717958647692 / mNumSides;

    if (distToGo > 1.0e-10) {

	double delta	      = atan(
				fabs(distToGo - sideLength * 0.5) /
				(circumRadius * sin(alpha)));
	param = distToGo > sideLength * 0.5? 
		((numSides + 0.5) * includedAngle + delta)  :
		((numSides + 0.5) * includedAngle - delta);

    } else {
	param = numSides * includedAngle;
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getDistAtPoint  (const AcGePoint3d& point,
                           double&            dist)  const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    double param;

    if ((es = this->getParamAtPoint(point, param)) != Acad::eOk) {
        return es;
    }
    if ((es = this->getDistAtParam(param, dist)) != Acad::eOk) {
        return es;
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getPointAtDist  (double             dist,
                           AcGePoint3d&       point) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    double param;

    if((es = this->getParamAtDist(dist, param)) != Acad::eOk) {
        return es;
    }
    if ((es = this->getPointAtParam(param, point)) != Acad::eOk) {
        return es;
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getFirstDeriv (double             param, 
                      AcGeVector3d&      firstDeriv) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }
   
    double includedAngle     = 6.28318530717958647692 / mNumSides;
    int    numIncludedAngles = param / includedAngle;
    double theta	     = param - numIncludedAngles * includedAngle;

    if (theta > 1.0e-10) {		// On a side

	firstDeriv = vertexArray[numIncludedAngles + 1] -
		     vertexArray[numIncludedAngles];
	firstDeriv.normalize();

    } else {				// On a vertex
	return Acad::eSingularPoint;
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getFirstDeriv (const AcGePoint3d& point, 
                      AcGeVector3d&      firstDeriv) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    double param;
    if ((es = this->getParamAtPoint(point, param)) != Acad::eOk) {
        return es;
    }

    return this->getFirstDeriv(param, firstDeriv);
}


Acad::ErrorStatus 
AcDbPoly::getClosestPointTo (const AcGePoint3d& givenPoint,
                             AcGePoint3d&       pointOnCurve,
                             Adesk::Boolean     /*extend*/) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeLineSeg3d lnsg(vertexArray[0], vertexArray[1]);
    double dist = lnsg.distanceTo(givenPoint);
    int index = 0;

    double dist2;
    for (int i = 1; i < vertexArray.length() - 1; i++) {

	lnsg.set(vertexArray[i], vertexArray[i + 1]);
	if ((dist2 = lnsg.distanceTo(givenPoint)) < dist) {
	    dist = dist2;
	    index = i;
	}
    }

    lnsg.set(vertexArray[index], vertexArray[index + 1]);
    pointOnCurve = lnsg.closestPointTo(givenPoint);

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getClosestPointTo (const AcGePoint3d&  givenPoint,
                             const AcGeVector3d& normal,
                             AcGePoint3d&        pointOnCurve,
                             Adesk::Boolean      /*extend*/) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Project the polygon as well as the point onto a plane
    // defined by the polygon center and the normal.

    AcGePlane plane(mCenter, normal);
    AcDbCurve* curve = NULL;
    if ((es = this->getOrthoProjectedCurve(plane, curve)) != Acad::eOk) {
        return es;
    }
    assert(curve != NULL);
    AcGePoint3d pt = givenPoint.orthoProject(plane);

    // Find the nearest point on the projected curve from the
    // projected point.

    AcGePoint3d newPt;
    if ((es = curve->getClosestPointTo(pt, newPt)) != Acad::eOk) {
        return es;
    }

    // Project the point back onto the polygon.
    //
    // Define a line starting from the newPt and along the normal.
    // Intersect the polygon with that line. Find all the points
    // and pick the one closest to the given point.

    AcGePoint3dArray pts;
    AcGeLine3d line(newPt, normal);
    if ((es = intLine(this, line, pts)) != Acad::eOk) {
        return es;
    }

    pointOnCurve = pts[0];
    double length = (givenPoint - pts[0]).length();
    double length2;
    for (int i = 1; i < pts.length(); i++) {
	if ((length2 = (givenPoint - pts[i]).length()) < length) {
	    pointOnCurve = pts[i];
	    length = length2;
	}
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getOrthoProjectedCurve(const AcGePlane&  plane,
                                 AcDbCurve*&       projectedCurve) const
{
    assertReadEnabled();

    AcGeVector3d projDir = plane.normal();
    return this->getProjectedCurve(plane, projDir, projectedCurve);
}


Acad::ErrorStatus 
AcDbPoly::getProjectedCurve(const AcGePlane&    plane,
                                 const AcGeVector3d& projDir,
                                 AcDbCurve*&         projectedCurve) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // project the start point and the center and create a new polygon.
    // The normal becomes the normal of the plane.

    if (projDir.isPerpendicularTo(plane.normal()))
	return Acad::eInvalidInput;
    
    if (projDir.isPerpendicularTo(mPlaneNormal)) {

	// Poly becomes a polyline.

        AcGePoint3dArray vertexArray;
        if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
            return es;
        }

	for (int i = 0; i < vertexArray.length(); i++) {
	    vertexArray[i] = vertexArray[i].project(plane, projDir);
	}
	AcDb2dPolyline* pline;
	if ((es = rx_makePline(vertexArray, pline)) != Acad::eOk) {
            return es;
        }
	pline->setPropertiesFrom(this);

	projectedCurve = pline;

    } else {

	// Poly becomes another poly.

	AcGePoint3d center, startPoint;
	center = mCenter.project(plane, projDir);
	startPoint = mStartPoint.project(plane, projDir);

	AcDbPoly* newPoly = new AcDbPoly(center, startPoint, 
					 mNumSides, plane.normal(), 
					 mpName);

	newPoly->setPropertiesFrom(this);
	projectedCurve = newPoly;
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getOffsetCurves(
    double offsetDist,
    AcDbVoidPtrArray& offsetCurves)  const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Offset distance is measured perpendicular to each side.

    double theta = 3.14159265358979323846 / mNumSides;
    double extendBy = offsetDist / cos(theta);

    AcGeVector3d x = mStartPoint - mCenter;
    x.normalize();
    x *= extendBy;

    AcGePoint3d startPt = mStartPoint + x;

    AcDbPoly* offsetPoly = new AcDbPoly(mCenter, startPt, mNumSides,
						 mPlaneNormal, mpName);
    assert(offsetPoly != NULL);
    offsetPoly->setPropertiesFrom(this);
    offsetCurves.append(offsetPoly);

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getSplitCurves(
    const AcGeDoubleArray& parameters,
    AcDbVoidPtrArray& curveSegments) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Since this is a closed periodic curve, the number of segemnts
    // that will result from splitting will be the same as the number
    // parameters we are splitting at. Each of the segments will be
    // a 2d polyline.

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    double param, nextParam, includedAngle = 6.28318530717958647692 / mNumSides;
    int    thisIndex, nextIndex;

    AcGePoint3dArray pts;
    AcGePoint3d   pt;

    AcGeMatrix3d  mat;
    this->getCoordSys(mat);

    AcDb2dPolyline* pline = NULL;

    for (int i = 0; i < parameters.length(); i++) {
	
	param = parameters[i];
	if (i < parameters.length() - 1) {
	    nextParam = parameters[i + 1];
	} else {
	    nextParam = parameters[0];
	}

	rx_fixangle(param);
	rx_fixangle(nextParam);

	thisIndex = param / includedAngle + 1;
	nextIndex = nextParam / includedAngle;

	double theta = nextParam - nextIndex * includedAngle;
	if (theta < 1.0e-10) {
	    
	    // Reduce nextIndex by one since we are going to evaluate
	    // at nextParam anyway. Since integer division truncates,
	    // we cannot decrement nextIndex all the time.

	    nextIndex -= 1;
	}

	if (nextIndex < thisIndex)
	    nextIndex += mNumSides;

	if ((es = this->getPointAtParam(param, pt)) != Acad::eOk) {
            return es;
        }
	pts.append(pt);

	int idx;
	for (int j = thisIndex; j <= nextIndex; j++) {

	    idx = j;
	    rx_fixindex(idx, mNumSides);
	    pts.append(vertexArray[idx]);
	}
	AOK(this->getPointAtParam(nextParam, pt));
	pts.append(pt);

	AOK(rx_makePline(pts, mat, pline));
	pline->setPropertiesFrom(this);
	pts.setPhysicalLength(0);
	curveSegments.append(pline);
    }

    return es;
}


Acad::ErrorStatus 
AcDbPoly::getSplitCurves(
    const AcGePoint3dArray& points,
    AcDbVoidPtrArray& curveSegments) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;
    AcGeDoubleArray params;
    params.setLogicalLength(points.length());
    for (int i = 0; i < points.length(); i++) {
	es = this->getParamAtPoint(points[i], params[i]);
    }

    return this->getSplitCurves(params, curveSegments);
}


Acad::ErrorStatus 
AcDbPoly::getArea	 (double&	       area) const
{
    assertReadEnabled();

    double circumRadius = (mStartPoint - mCenter).length();

    area = mNumSides * circumRadius * cos(3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides)
		     * circumRadius * sin(3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides);

    return Acad::eOk;
}


//*************************************************************************
// Overridden methods from AcDbEntity
//*************************************************************************


Acad::ErrorStatus 
AcDbPoly::getSubentPathsAtGsMarker(AcDb::SubentType      type, 
                       int                   gsMark,
                       const AcGePoint3d&    pickPoint,
                       const AcGeMatrix3d&   viewXform,
                       int&                  numPaths, 
                       AcDbFullSubentPath*&  subentPaths,
                       int                   numInserts,
                       AcDbObjectId*         entAndInsertStack) const
{
    assertReadEnabled();

    assert(numInserts == 0);
    assert(entAndInsertStack == NULL);

    numPaths    = 0;
    subentPaths = NULL;
    
    if (gsMark == 0)
	return eInvalidInput;

    switch (type) {
	
    case kVertexSubentType:

	if (gsMark > 0) {
	    numPaths       = 2;
	    subentPaths    = new AcDbFullSubentPath[2];
	    subentPaths[0] = AcDbFullSubentPath(objectId(), 
                                              kVertexSubentType, gsMark - 1);
	    subentPaths[1] = AcDbFullSubentPath(objectId(), 
                                              kVertexSubentType, gsMark);
	} else {
	    assert(kFalse);
	}
	
    case kEdgeSubentType:
	
	if (gsMark > 0) {
	    numPaths       = 1;
	    subentPaths    = new AcDbFullSubentPath[1];
	    subentPaths[0] = AcDbFullSubentPath(objectId(), 
                                              kEdgeSubentType, gsMark - 1);
	} else {
	    assert(kFalse);
	}
	
	break;
	
    default:
	assert(kFalse);
    }
    return eOk;
}


Acad::ErrorStatus 
AcDbPoly::getGsMarkersAtSubentPath(const AcDbFullSubentPath& path, 
                              AcDbIntArray&         gsMarkers) const
{
    Acad::ErrorStatus es;

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcDbSubentId subEntId = path.subentId();

    switch (subEntId.type()) {
	
    case kVertexSubentType:

	if (subEntId.index() == 0) 
	    gsMarkers.append(subEntId.index() + 1);
	else
	    gsMarkers.append(subEntId.index());
	break;

    case kEdgeSubentType:

	gsMarkers.append(subEntId.index() + 1);
	break;
	
    default:
	assert(kFalse);
    }

    return eOk;
}


AcDbEntity* 
AcDbPoly::subentPtr(const AcDbFullSubentPath& path) const
{
    Acad::ErrorStatus es;

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return NULL;
    }

    AcDbSubentId subEntId = path.subentId();
    AcDbPoint *pt;
    AcDbLine *line;

    switch (subEntId.type()) {
	
    case kVertexSubentType:

	pt = new AcDbPoint();
	pt->setPosition(vertexArray[subEntId.index()]);
	pt->setNormal(this->normal());
	pt->setPropertiesFrom(this);
	return pt;

    case kEdgeSubentType:

	line = new AcDbLine();
	line->setStartPoint(vertexArray[subEntId.index()]);
	line->setEndPoint(vertexArray[subEntId.index() + 1]);
	line->setNormal(this->normal());
	line->setPropertiesFrom(this);
	return line;
	
    default:
	assert(kFalse);
    }

    return NULL;
}

Acad::ErrorStatus   
AcDbPoly::getOsnapPoints(
                         AcDb::OsnapMode       osnapMode,
                         int                   gsSelectionMark,
                         const AcGePoint3d&    pickPoint,
                         const AcGePoint3d&    lastPoint,
                         const AcGeMatrix3d&   viewXform,
  	                 AcGePoint3dArray&        snapPoints,
                         AcDbIntArray&     /*geomIds*/) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    if (gsSelectionMark == 0)
	return Acad::eOk;

    Adesk::Boolean unknownMode = (osnapMode != AcDb::kOsModeEnd  && 
				  osnapMode != AcDb::kOsModeMid  &&
				  osnapMode != AcDb::kOsModeNear && 
				  osnapMode != AcDb::kOsModePerp &&
				  osnapMode != AcDb::kOsModeCen);
    if (unknownMode)
	return Acad::eOk;

    int startIndex = gsSelectionMark - 1;

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeLineSeg3d lnsg(vertexArray[startIndex], vertexArray[startIndex + 1]);
    AcGePoint3d  pt;

    AcGeLine3d line, perpLine;
    AcGeVector3d vec;

    AcGeVector3d viewDir(viewXform(Z, 0), viewXform(Z, 1), viewXform(Z, 2));

    switch (osnapMode) {

    case AcDb::kOsModeEnd:
	
	snapPoints.append(vertexArray[startIndex]); 
	snapPoints.append(vertexArray[startIndex + 1]); 
	break;

    case AcDb::kOsModeMid:

	pt.set(
	((vertexArray[startIndex])[X] + (vertexArray[startIndex + 1])[X]) * 0.5,
	((vertexArray[startIndex])[Y] + (vertexArray[startIndex + 1])[Y]) * 0.5,
	((vertexArray[startIndex])[Z] + (vertexArray[startIndex + 1])[Z]) * 0.5
	);
	snapPoints.append(pt);
	break;

    case AcDb::kOsModeNear:

	vec = vertexArray[startIndex + 1] - vertexArray[startIndex]; 
	vec.normalize();
	line.set(vertexArray[startIndex], vec);
	pt = line.projClosestPointTo(pickPoint, viewDir);
	snapPoints.append(pt);
	break;

    case AcDb::kOsModePerp:

	// create a line (semi-infinite) and find a point on it.

	vec = vertexArray[startIndex + 1] - vertexArray[startIndex]; 
	vec.normalize();
	line.set(vertexArray[startIndex], vec);
	pt = line.closestPointTo(lastPoint);
	snapPoints.append(pt);
	break;

    case AcDb::kOsModeCen:

	snapPoints.append(mCenter);
	break;

    default:
	return Acad::eOk;

    }

    return es;
}


Acad::ErrorStatus   
AcDbPoly::getGripPoints(AcGePoint3dArray&     gripPoints,
                        AcDbIntArray&  osnapModes,
                        AcDbIntArray&  geomIds) const
{
    assertReadEnabled();

    Acad::ErrorStatus es;

    if ((es = this->getVertices(gripPoints)) != Acad::eOk) {
        return es;
    }

    // Remove the duplicate point at the start/end and add
    // center as the last point.

    gripPoints.removeAt(gripPoints.length() - 1);

    // Figure out what this geomId really means!!! TODO.

    for (int i = 0; i < gripPoints.length(); i++) {
        osnapModes.append(kOsModeEnd);
        geomIds.append(i + 1);
    }

    gripPoints.append(mCenter);
    osnapModes.append(kOsModeCen);
    geomIds.append(gripPoints.length());

    return es;
}


Acad::ErrorStatus   
AcDbPoly::moveGripPointsAt(const AcDbIntArray& indices,
	                   const AcGeVector3d& offset)
{
    assertWriteEnabled();

    Acad::ErrorStatus es = Acad::eOk;
    AcDbIntArray move;

    int numGripPoints = mNumSides + 1;

    move.setLogicalLength(numGripPoints);
    move.setAll(kFalse);
    for (int j = 0; j < indices.length(); j++) {
	if (indices[j] < numGripPoints) {
	    move[indices[j]] = kTrue;
	}
    }

    for (int i = 0; i < numGripPoints; i++) {

	if (move[i]) {

	    // If one of the vertices of the polygon is moved, find out
	    // the new start point from that movement and set it.
	    //
	    // One way to do that would be to define a new polygon taking
	    // the moved point as the start point. Get the vertices of that
	    // polygon, and the start point of the old polygon would be
	    // one of the vertices of the new polygon at the appropriate
	    // index.

	    // If the center is moved, just translate the polygon.

	    if (i < mNumSides) {  // vertex movement

		AcGePoint3dArray oldVertexArray;
		AOK(this->getVertices(oldVertexArray));

		AcGePoint3d vertex = oldVertexArray[i];
		vertex += offset;

		if (vertex == mCenter) {
		    return Acad::eOk;
		}

		AcDbPoly* newPoly = new AcDbPoly(mCenter, vertex,
						 mNumSides, mPlaneNormal,
						 mpName);
		AcGePoint3dArray newVertexArray;
		AOK(newPoly->getVertices(newVertexArray));

		delete newPoly;

		mStartPoint = newVertexArray[mNumSides - i];

	    } else {		 // center movement

		AcGeMatrix3d mat;

		mat.entry[0][3] = offset[X];
		mat.entry[1][3] = offset[Y];
		mat.entry[2][3] = offset[Z];

		this->transformBy(mat);
	    }
	}
    }

    return es;
}


Acad::ErrorStatus	
AcDbPoly::getStretchPointsAt(AcGePoint3dArray&  stretchPoints) const
{
    // Stretch points are the same as grip points except for the
    // center of the polygon.

    assertReadEnabled();

    Acad::ErrorStatus es;

    if ((es = this->getVertices(stretchPoints)) != Acad::eOk) {
        return es;
    }

    // Remove the duplicate point at the start/end.

    stretchPoints.removeAt(stretchPoints.length() - 1);

    return es;
}


Acad::ErrorStatus         
AcDbPoly::moveStretchPointsAt(const AcDbIntArray& indices,
                              const AcGeVector3d&     offset)
{
    assertWriteEnabled();

    Acad::ErrorStatus es = eOk;

    if (indices.length() == 0 || offset.isZeroLength()) {
	return eOk;
    }
    
    // If all the stretch points are selected call transfomBy(), 
    // otherwise call moveGripPoints().

    Boolean move;

    if (indices.length() < mNumSides) {
	move = kFalse;
    } else {
	move = kTrue;
    }

    if (move) {

	AcGeMatrix3d mat;

	mat(0, 3) = offset[0];
	mat(1, 3) = offset[1];
	mat(2, 3) = offset[2];

	return transformBy(mat);

    } else {
	return moveGripPointsAt(indices, offset);
    }
}


Acad::ErrorStatus AcDbPoly::getTransformedCopy(const AcGeMatrix3d& mat,
                                               AcDbEntity*&        ent) const
{
    assertReadEnabled();
    Acad::ErrorStatus es = eOk;

    // Transform the points and create a polyline out of them.

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    for (int i = 0; i < vertexArray.length(); i++) {
	vertexArray[i].transformBy(mat);
    }

    AcDb3dPolyline *pline = NULL;
    if ((es = rx_makePline(vertexArray, pline)) != Acad::eOk) {
        return es;
    }
    assert(pline != NULL);
    pline->setPropertiesFrom(this);

    ent = pline;

    return es;
}


Acad::ErrorStatus   
AcDbPoly::intersectWith(const AcDbEntity*   ent,
			     AcDb::Intersect     intType,
			     AcGePoint3dArray&      points,
			     int                 /*thisGsMarker*/,
			     int                 /*otherGsMarker*/) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    assert(ent != NULL);

    // The idea is to intersect each side of the polygon with the
    // given entity and return all the points.
    //
    // For non-R12-entities, i.e. the ones that have their intersection
    // methods defined, we call that method for each of the sides of the
    // polygon. For R12-entities, we use the locally defined intersectors
    // since their protocols are not implemented.

    if (ent->isKindOf(AcDbLine::desc())) {

	if ((es = intLine(this, AcDbLine::cast(ent), intType, NULL, points))
               != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDbArc::desc())) {

	if((es = intArc(this, AcDbArc::cast(ent), intType, NULL, points)) 
              != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDbCircle::desc())) {

	if ((es = intCircle(this, AcDbCircle::cast(ent), intType, NULL, points))
               != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDb2dPolyline::desc())) {

	if ((es = intPline(this, AcDb2dPolyline::cast(ent), intType, 
                           NULL, points)) != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDb3dPolyline::desc())) {

	if ((es = intPline(this, AcDb3dPolyline::cast(ent), intType, 
                           NULL, points)) != Acad::eOk) {
            return es;
        }

    } else {

        AcGePoint3dArray vertexArray;
        if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
            return es;
        }

	if (intType == kExtendArg || intType == kExtendBoth) {
	    intType = kExtendThis;
	}

	AcDbLine* acadLine;

	for (int i = 0; i < vertexArray.length() - 1; i++) {

	    acadLine = new AcDbLine();
	    acadLine->setStartPoint(vertexArray[i]);
	    acadLine->setEndPoint(vertexArray[i + 1]);
	    acadLine->setNormal(this->normal());

	    if ((es = ent->intersectWith(acadLine, intType, points))
                   != Acad::eOk) {
                delete acadLine;
                return es;
            }

	    delete acadLine;
	}
    }

    return es;
}

 
Acad::ErrorStatus   
AcDbPoly::intersectWith( const AcDbEntity*   ent,
			      AcDb::Intersect     intType,
			      const AcGePlane&    projPlane,
			      AcGePoint3dArray&      points,
			      int                 /*thisGsMarker*/,
			      int                 /*otherGsMarker*/) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    assert(ent != NULL);

    // The idea is to intersect each side of the polygon with the
    // given entity and return all the points.
    //
    // For non-R12-entities, i.e. the ones that have their intersection
    // methods defined, we call that method for each of the sides of the
    // polygon. For R12-entities, we use the locally defined intersectors
    // since their protocols are not implemented.

    if (ent->isKindOf(AcDbLine::desc())) {

	if((es = intLine(this, AcDbLine::cast(ent), intType, 
                         &projPlane, points)) != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDbArc::desc())) {

	if ((es = intArc(this, AcDbArc::cast(ent), intType, &projPlane, points))
               != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDbCircle::desc())) {

	if ((es = intCircle(this, AcDbCircle::cast(ent), intType, &projPlane, 
                            points)) != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDb2dPolyline::desc())) {

	if ((es = intPline(this, AcDb2dPolyline::cast(ent), intType, &projPlane,
                           points)) != Acad::eOk) {
            return es;
        }

    } else if (ent->isKindOf(AcDb3dPolyline::desc())) {

	if ((es = intPline(this, AcDb3dPolyline::cast(ent), intType, &projPlane,
                           points)) != Acad::eOk) {
            return es;
        }

    } else {

        AcGePoint3dArray vertexArray;
        if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
            return es;
        }

	if (intType == kExtendArg || intType == kExtendBoth) {
	    intType = kExtendThis;
	}

	AcDbLine* acadLine;

	for (int i = 0; i < vertexArray.length() - 1; i++) {

	    acadLine = new AcDbLine();
	    acadLine->setStartPoint(vertexArray[i]);
	    acadLine->setEndPoint(vertexArray[i + 1]);
	    acadLine->setNormal(this->normal());

	    if ((es = ent->intersectWith(acadLine, 
                                              intType, 
                                              projPlane, 
                                              points)) != Acad::eOk) {
                delete acadLine;
                return es;
            }

	    delete acadLine;
	}

	// All the points that we got in this process are on the other curve
	// and we are dealing with apparent intersection. If the other curve
	// is 3D or is not on the same plane as poly, the points are not on
	// poly. 
	//
	// So we need to do some more work. Project the points back onto the
	// plane. They should lie on the projected poly. Find points on real 
	// poly corresponding to the projected points.

        AcGePoint3d    projPt, planePt;
        AcGePoint3dArray  pts;
        AcGeLine3d     line;
	AcGePlane      polyPlane(mCenter, mPlaneNormal);

	for (i = 0; i < points.length(); i++) {

	    // Define a line starting from the projPt and along the normal.
	    // Intersect the polygon with that line. Find all the points
	    // and pick the one closest to the given point.

	    projPt = points[i].orthoProject(projPlane);
	    line.set(projPt, projPlane.normal());
	    if ((es = intLine(this, line, pts)) != Acad::eOk) {
                return es;
            }

	    planePt = projPt.project(polyPlane, projPlane.normal());

	    points[i] = pts[0];
	    double length = (planePt - pts[0]).length();
	    double length2;

	    for (int j = 1; j < pts.length(); j++) {
		if ((length2 = (planePt - pts[j]).length()) < length) {
		    points[i] = pts[j];
		    length = length2;
		}
	    }
	}
    }

    return es;
}


void		
AcDbPoly::list() const
{
    assertReadEnabled();

    AcDbEntity::list();

    double circumRadius = (mStartPoint - mCenter).length();
    double sideLength = 2.0 * circumRadius * cos(3.14159265358979323846 * 0.5 - 3.14159265358979323846 / mNumSides);

    ads_printf("%18s%16s %d \n", /*MSG0*/"",
                          "Number Of Sides:", mNumSides);

    ads_printf("%18s%16s %-9.16q0 \n", /*MSG0*/"",
                          "Circumference:", sideLength * mNumSides);

    double areaVal;
    this->getArea(areaVal);
    ads_printf("%18s%16s %-9.16q0 \n", /*MSG0*/"", "Area:", areaVal);
    
    ads_printf("%18s%16s ", /*MSG0*/"", "Center:");

    ads_point center;
    center[0] = mCenter[0]; center[1] = mCenter[1]; center[2] = mCenter[2];
    rx_wc2uc(center, center, Adesk::kFalse);
    ads_printf( "X = %-9.16q0, Y = %-9.16q0, Z = %-9.16q0\n", 
		      center[0], center[1], center[2]);

    ads_point startPt;
    startPt[0] = mStartPoint[0]; 
    startPt[1] = mStartPoint[1]; 
    startPt[2] = mStartPoint[2];
    rx_wc2uc(startPt, startPt, Adesk::kFalse);
    ads_printf("%18s%16s ", /*MSG0*/"", "Start Point:");

    ads_printf( "X = %-9.16q0, Y = %-9.16q0, Z = %-9.16q0\n", 
		      startPt[0], startPt[1], startPt[2]);

    ads_point normal;
    normal[0] = mPlaneNormal[0]; 
    normal[1] = mPlaneNormal[1]; 
    normal[2] = mPlaneNormal[2];
    rx_wc2uc(normal, normal, Adesk::kFalse);
    ads_printf("%18s%16s ", /*MSG0*/"", "Normal:");

    ads_printf( "X = %-9.16q0, Y = %-9.16q0, Z = %-9.16q0\n", 
		      normal[0], normal[1], normal[2]);

}


Acad::ErrorStatus   
AcDbPoly::transformBy(const AcGeMatrix3d& xform)
{
    assertWriteEnabled();

    mCenter.	 transformBy(xform);
    mStartPoint. transformBy(xform);

    mPlaneNormal.transformBy(xform);
    mPlaneNormal.normalize();

    return Acad::eOk;
}


Acad::ErrorStatus	
AcDbPoly::explode(AcDbVoidPtrArray& entitySet) const
{
    assertReadEnabled();

    Acad::ErrorStatus es = Acad::eOk;

    // Exploding a polygon produces a set of lines which
    // are the sides of the polygon.

    AcGePoint3dArray vertexArray;
    if ((es = this->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcDbLine* line;

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	line = new AcDbLine();
	line->setStartPoint(vertexArray[i]);
	line->setEndPoint(vertexArray[i + 1]);
	line->setNormal(mPlaneNormal);

	entitySet.append(line);
    }

    return es;
}


Adesk::Boolean 
AcDbPoly::worldDraw(AcGiWorldDraw*	  worldDraw)
{
    assertReadEnabled();

    AcDbEntity::worldDraw(worldDraw);

    if (worldDraw->regenAbort()) {
	return Adesk::kTrue;
    }

    AOK(drawName(this, worldDraw));

    switch (worldDraw->regenType()) {
	
    case kAcGiHideOrShadeCommand:
    case kAcGiRenderCommand:
	
	AOK(drawFacets(this, worldDraw, NULL));
	AOK(drawEdges (this, worldDraw, NULL));
	break;
	
    case kAcGiStandardDisplay:
    case kAcGiSaveWorldDrawForR12:
	
	AOK(drawEdges (this, worldDraw, NULL));
	break;
	
    default:
	assert(Adesk::kFalse);

    }
    
    return Adesk::kTrue; // Don't call viewportDraw().
}


void		
AcDbPoly::viewportDraw(AcGiViewportDraw*  vportDraw)
{
    assertReadEnabled();

    AcDbEntity::viewportDraw(vportDraw);

    if (vportDraw->regenAbort()) {
	return;
    }

    switch (vportDraw->regenType()) {
	
    case kAcGiHideOrShadeCommand:
    case kAcGiRenderCommand:
	
	AOK(drawFacets(this, NULL, vportDraw));
	AOK(drawEdges (this, NULL, vportDraw));
	break;
	
    case kAcGiStandardDisplay:
    case kAcGiSaveWorldDrawForR12:
	
	AOK(drawEdges (this, NULL, vportDraw));
	break;
	
    default:
	assert(Adesk::kFalse);

    }
}



//*************************************************************************
// Overridden methods from AcDbObject
//*************************************************************************


Acad::ErrorStatus 
AcDbPoly::applyPartialUndo(AcDbDwgFiler* filer,
			   AcRxClass*    classObj)
{
    // First thing we do is see if the class matches with ours. 
    // If it doesn't, we call base class's applyPartialUndo(),
    // and, hopefully, one of them will take care of it.

    if (classObj != AcDbPoly::desc())
	return AcDbCurve::applyPartialUndo(filer, classObj);

    // Read the op-code and call the appropriate "set" method to 
    // undo what was done. The "set" does the filing again for redo.

    PolyOpCodeForPartialUndo code;
    Adesk::Int16	 shortCode;
    filer->readItem(&shortCode);
    code = (PolyOpCodeForPartialUndo)shortCode;

    Adesk::UInt32	value32;
 
    switch (code) {

    case kSetNumSides:
	filer->readItem(&value32);
	AOK(setNumSides(value32));
	break;

    default:
	assert(kFalse);
	break;
    }

    return eOk;
}


Acad::ErrorStatus   
AcDbPoly::dwgInFields(AcDbDwgFiler* filer)
{
    assertWriteEnabled();

    Acad::ErrorStatus es;

    if ((es = AcDbEntity::dwgInFields(filer)) != Acad::eOk) {
        return es;
    }

    filer->readItem(&mCenter);
    filer->readItem(&mStartPoint);
    Adesk::Int32 int32;
    filer->readItem(&int32);
    mNumSides = int32;
    filer->readItem(&mPlaneNormal);
    filer->readItem(&mpName);
    filer->readItem(&mTextStyle);

    return es;
}


Acad::ErrorStatus   
AcDbPoly::dwgOutFields(AcDbDwgFiler* filer) const
{
    assertReadEnabled();

    Acad::ErrorStatus es;

    if ((es = AcDbEntity::dwgOutFields(filer)) != Acad::eOk) {
        return es;
    }

    filer->writeItem(mCenter);
    filer->writeItem(mStartPoint);
    Adesk::Int32 int32 = mNumSides;
    filer->writeItem(int32);
    filer->writeItem(mPlaneNormal);
    filer->writeItem(mpName);

    // mTextStyle is a hard pointer id, so filing it out to
    // the purge filer (kPurgeFiler) prevents purging of this
    // object.
    filer->writeItem(mTextStyle);

    return es;
}


Acad::ErrorStatus   
AcDbPoly::dxfInFields(AcDbDxfFiler* filer)
{
    assertWriteEnabled();

    Acad::ErrorStatus  es = eOk;
    resbuf	       rb;

    if ( (AcDbEntity::dxfInFields(filer) != Acad::eOk) ||
         !filer->atSubclassData("AcDbPoly") )
    {
	return filer->filerStatus();
    }

    while ((es == eOk) && ((es = filer->readResBuf(&rb)) == eOk)) {
	
	switch (rb.restype) {
	    
	case kDxfXCoord:

	    mCenter    = asPnt3d(rb.resval.rpoint);
	    break;
	    
	case kDxfXCoord + 1:
	    
	    mStartPoint   = asPnt3d(rb.resval.rpoint);
	    break;
	    
	case kDxfInt32:
	    
	    mNumSides	 = rb.resval.rlong;
	    break;
	    
	case kDxfNormalX:
	    
	    mPlaneNormal = asVec3d(rb.resval.rpoint);
	    break;

	case kDxfText:

	    if (mpName != NULL)
		free(mpName);
	    mpName = strdup(rb.resval.rstring);
	    break;
	    
	case kDxfHardPointerId:

	    mTextStyle.setFromOldId(rb.resval.rlong);
	    break;
	    
	default:
	    // An unrecognized group. Push it back so that the subclass
	    // can read it again.
	    
	    filer->pushBackItem();
	    es = eEndOfFile;
	    break;
	} /*switch*/
    } /*while*/
 
	    
    // At this point the es variable must contain eEndOfFile - either 
    // from readResBuf() or from pushback. If not, it indicates that an 
    // error happened and we should return immediately.

    if (es != eEndOfFile)
	return eInvalidResBuf;

    return es;
}


Acad::ErrorStatus   
AcDbPoly::dxfOutFields(AcDbDxfFiler* filer) const
{
    assertReadEnabled();

    Acad::ErrorStatus es;

    if ((es = AcDbEntity::dxfOutFields(filer)) != Acad::eOk) {
        return es;
    }

    filer->writeItem(AcDb::kDxfSubclass, "AcDbPoly");      

    filer->writeItem(AcDb::kDxfXCoord, mCenter);
    filer->writeItem(AcDb::kDxfXCoord + 1, mStartPoint);
    Adesk::Int32 int32 = mNumSides;
    filer->writeItem(AcDb::kDxfInt32, int32);
    filer->writeItem(AcDb::kDxfNormalX, mPlaneNormal);
    filer->writeItem(AcDb::kDxfText, mpName);
    filer->writeItem(AcDb::kDxfHardPointerId, mTextStyle);

    return es;
}


//*************************************************************************
// Statics functions used in this file. 
//*************************************************************************

static Acad::ErrorStatus drawFacets(const AcDbPoly*    poly,
				    AcGiWorldDraw*     worldDraw,
				    AcGiViewportDraw*  vportDraw)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }
	
    // Since GS interface handles polygons with more than three sides,
    // we will pass in the whole thing. No need of triangulation.

    Adesk::UInt32 faceListSize = vertexArray.length() + 1; // one face + 
							   // num Vertices for
							   // the face.
    Adesk::Int32* faceList = new Adesk::Int32[faceListSize];

    faceList[0] = vertexArray.length();
    for (int i = 0; i < vertexArray.length(); i++) {
	faceList[i + 1] = i;
    }

    // If we wanted to do anything special for edges and faces, e.g. making
    // edges invisible, setting face color, etc., we would have bothered
    // with these two objects.

    AcGiEdgeData* edgeData = NULL;
    AcGiFaceData* faceData = NULL;
    
    if (worldDraw != NULL) {
	
	worldDraw->geometry().shell(vertexArray.length(),
				    vertexArray.asArrayPtr(),
				    faceListSize,
				    faceList,
				    edgeData,
				    faceData);
    } else {
	assert(Adesk::kFalse);
    }
    
    delete [] faceList;
    delete    edgeData;
    delete    faceData;

    return es;
}


static Acad::ErrorStatus drawEdges(const AcDbPoly*    poly,
				   AcGiWorldDraw*     worldDraw,
				   AcGiViewportDraw*  vportDraw)
{
    Acad::ErrorStatus es = Acad::eOk;

    // Draw each edge of the polygon as a line. We could have drawn
    // the whole polygon as a polyline, but we want to attach subEntity
    // traits (e.g. which line it is) to each line which will be used
    // for snap.
    //
    // Since we are drawing the polygon line by line, we also have the
    // control over setting the linetype and color of each line (via
    // subEntity traits), but we won't get that fancy.


    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGePoint3d ptArray[2];

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	if (worldDraw != NULL) {
	    worldDraw->subEntityTraits().setSelectionMarker(i + 1);
	} else {
	    assert(Adesk::kFalse);
	    //vportDraw->subEntityTraits().setSelectionMarker(i + 1);
	}

	ptArray[0] = vertexArray[i];
	ptArray[1] = vertexArray[i + 1];

	if (worldDraw != NULL) {
	    worldDraw->geometry().polyline(2, ptArray);
	} else {
	    assert(Adesk::kFalse);
	    //vportDraw->geometry().polyline3d(2, ptArray);
	}
    }

    return es;
}


static Acad::ErrorStatus drawName(const AcDbPoly*    poly,
				  AcGiWorldDraw*     worldDraw)
{
    char *pName = poly->name();
    AcDbObjectId styleId = poly->styleId();

    AcGiTextStyle textStyle;

    if (styleId != NULL)
        if (rx_getTextStyle(textStyle, styleId) != eOk)
            styleId = AcDbObjectId::kNull;

    if (pName != NULL) {
	AcGeVector3d direction = poly->startPoint() - poly->center();
	AcGePoint3d center = poly->center();
	AcGeVector3d normal = poly->normal();
	Adesk::Boolean ok;

        if (styleId)
	    ok = 
	        worldDraw->geometry().text(center, normal, direction,
                 pName, -1, 0, textStyle);
        else
	    ok = 
	        worldDraw->geometry().text(center, normal, direction,
                 direction.length() / 20, 1, 0, pName);

	free(pName);
    }
    return Acad::eOk;
}


static Acad::ErrorStatus intLine(const AcDbPoly*      poly, 
				 const AcGeLine3d     line,
				 AcGePoint3dArray&       points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeLineSeg3d tlnsg;
    AcGePoint3d   pt;

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	tlnsg.set(vertexArray[i], vertexArray[i + 1]);

	if (!tlnsg.intersectWith(line, pt)) {
	    continue;
	} else {
	    points.append(pt);
	}
    }

    return es;
}


static Acad::ErrorStatus intLine(const AcDbPoly*   poly, 
				 const AcDbLine*   line,
				 AcDb::Intersect   intType,
				 const AcGePlane*  projPlane,
				 AcGePoint3dArray&    points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGeLineSeg3d lnsg(line->startPoint(), line->endPoint());
    es = intLine(poly, lnsg, intType, projPlane, points);

    return es;
}


static Acad::ErrorStatus intLine(const AcDbPoly*      poly, 
				 const AcGeLineSeg3d  lnsg,
				 AcDb::Intersect      intType,
				 const AcGePlane*     projPlane,
				 AcGePoint3dArray&       points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeLine3d aline(lnsg.startPoint(), lnsg.endPoint());
    AcGeLineSeg3d tlnsg;
    AcGePoint3d   pt;
    AcGePoint3d   dummy;

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	tlnsg.set(vertexArray[i], vertexArray[i + 1]);

	if (intType == AcDb::kExtendArg || intType == AcDb::kExtendBoth) {
	    if (projPlane == NULL) {
		if (!tlnsg.intersectWith(aline, pt)) {
		    continue;
		} else {
		    points.append(pt);
		}
	    } else {
		if (!tlnsg.projIntersectWith(aline, projPlane->normal(),
		    pt, dummy))
		{
		    continue;
		} else {
		    points.append(pt);
		}
	    }
	} else {
	    if (projPlane == NULL) {
		if (!tlnsg.intersectWith(lnsg, pt)) {
		    continue;
		} else {
		    points.append(pt);
		}
	    } else {
		if (!tlnsg.projIntersectWith(lnsg, projPlane->normal(),
		    pt, dummy))
		{
		    continue;
		} else {
		    points.append(pt);
		}
	    }
	}
    }

    return es;
}

static Acad::ErrorStatus intArc(const AcDbPoly*  poly, 
				const AcDbArc*	 arc,
				AcDb::Intersect  intType,
				const AcGePlane* projPlane,
				AcGePoint3dArray&   points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGeCircArc3d aarc(arc->center(), arc->normal(), 
                       arc->normal().perpVector(), arc->radius(), 
                       arc->startAngle(), arc->endAngle());
    es = intArc(poly, aarc, intType, projPlane, points);

    return es;
}


static Acad::ErrorStatus intArc(const AcDbPoly*       poly, 
				const AcGeCircArc3d   arc,
				AcDb::Intersect       intType,
				 const AcGePlane*     projPlane,
				AcGePoint3dArray&        points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeCircArc3d  acircle(arc.center(), arc.normal(), arc.radius());
    AcGeLineSeg3d lnsg;
    AcGePoint3d   pt1, pt2;
    AcGePoint3d   dummy1, dummy2;
    int           howMany;

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	lnsg.set(vertexArray[i], vertexArray[i + 1]);

	if (intType == AcDb::kExtendArg || intType == AcDb::kExtendBoth) {
	    if (projPlane == NULL) {
		if (!acircle.intersectWith(lnsg, howMany, pt1, pt2)) {
		    continue;
		} else {
		    if (howMany > 1) {
			points.append(pt1);
			points.append(pt2);
		    } else {
			points.append(pt1);
		    }
		}
	    } else {
		if (!acircle.projIntersectWith(lnsg, projPlane->normal(),
		    howMany, pt1, pt2, dummy1, dummy2))
		{
		    continue;
		} else {
		    if (howMany > 1) {
			points.append(pt1);
			points.append(pt2);
		    } else {
			points.append(pt1);
		    }
		}
	    }
	} else {
	    if (projPlane == NULL) {
		if (!arc.intersectWith(lnsg, howMany, pt1, pt2)) {
		    continue;
		} else {
		    if (howMany > 1) {
			points.append(pt1);
			points.append(pt2);
		    } else {
			points.append(pt1);
		    }
		}
	    } else {
		if (!arc.projIntersectWith(lnsg, projPlane->normal(),
		    howMany, pt1, pt2, dummy1, dummy2))
		{
		    continue;
		} else {
		    if (howMany > 1) {
			points.append(pt1);
			points.append(pt2);
		    } else {
			points.append(pt1);
		    }
		}
	    }
	}
    }

    return es;
}


static Acad::ErrorStatus intCircle(const AcDbPoly*    poly, 
				   const AcDbCircle*  circle,
				   AcDb::Intersect    intType,
				   const AcGePlane*   projPlane,
				   AcGePoint3dArray&     points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGeCircArc3d  acircle(circle->center(), circle->normal(),
		          circle->radius());
    es = intCircle(poly, acircle, intType, projPlane, points);

    return es;
}


static Acad::ErrorStatus intCircle(const AcDbPoly*      poly, 
				   const AcGeCircArc3d   circle,
				   AcDb::Intersect      intType,
				   const AcGePlane*     projPlane,
				   AcGePoint3dArray&       points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcGeLineSeg3d lnsg;
    AcGePoint3d   pt1, pt2;
    AcGePoint3d   dummy1, dummy2;
    int           howMany;

    for (int i = 0; i < vertexArray.length() - 1; i++) {

	lnsg.set(vertexArray[i], vertexArray[i + 1]);

	if (projPlane == NULL) {
	    if (!circle.intersectWith(lnsg, howMany, pt1, pt2)) {
		continue;
	    } else {
		if (howMany > 1) {
		    points.append(pt1);
		    points.append(pt2);
		} else {
		    points.append(pt1);
		}
	    }
	} else {
	    if (!circle.projIntersectWith(lnsg, projPlane->normal(), 
	       howMany, pt1, pt2, dummy1, dummy2))
	    {
		continue;
	    } else {
		if (howMany > 1) {
		    points.append(pt1);
		    points.append(pt2);
		} else {
		    points.append(pt1);
		}
	    }
	}
    }

    return es;
}


static Acad::ErrorStatus intPline(const AcDbPoly*         poly, 
				  AcDb2dPolyline*         pline,
				  AcDb::Intersect         intType,
				  const AcGePlane*        projPlane,
				  AcGePoint3dArray&          points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcDbSpline*      spline = NULL;
    AcDbLine*	     acadLine = NULL;

    AcGePoint3dArray    pts;
    AcGeDoubleArray  bulges;
    int              numPoints, i;
    AcGePoint3d      pt1, pt2;

    AcGeCircArc3d    arc;
    AcGeLineSeg3d    lnsg;

    AcGeVector3d entNorm = pline->normal();
    AcDb::Intersect type = kOnBothOperands;

    switch (pline->polyType()) {

    case k2dSimplePoly:
    case k2dFitCurvePoly:

	// Intersect with each line or arc segment of the polyline. 
	// Depending on the intType, extend the last segment.

	if ((es = rx_scanPline(pline, pts, bulges)) != Acad::eOk) {
            return es;
        }
	numPoints = pts.length();

	for (i = 0; i < numPoints - 1; i++) {

	    pt1 = pts[i]; pt2 = pts[i + 1];
	    if (i == numPoints - 2)
		type = intType;

	    if (bulges[i] > 1.0e-10) {     // create an arc

		rx_wc2ec(asDblArray(pt1), asDblArray(pt1), asDblArray(entNorm), 
							      Adesk::kFalse);
		rx_wc2ec(asDblArray(pt2), asDblArray(pt2), asDblArray(entNorm), 
							      Adesk::kFalse);

		AOK(rx_makeArc(pt1, pt2, bulges[i], entNorm, arc));
		AOK(intArc(poly, arc, type, projPlane, points));

	    } else {			   // create a line

		lnsg.set(pt1, pt2);
		AOK(intLine(poly, lnsg, type, projPlane, points));
	    }
	}
	break;

    case k2dQuadSplinePoly:
    case k2dCubicSplinePoly:

	if ((es = pline->getSpline(spline)) != Acad::eOk) {
            return es;
        }

	if (intType == kExtendArg || intType == kExtendBoth) {
	    intType = kExtendThis;
	}

	for (i = 0; i < vertexArray.length() - 1; i++) {

	    acadLine = new AcDbLine();
	    acadLine->setStartPoint(vertexArray[i]);
	    acadLine->setEndPoint(vertexArray[i + 1]);
	    acadLine->setNormal(poly->normal());

	    if (projPlane == NULL) {
		AOK(spline->intersectWith(acadLine, intType, points));
	    } else {
		AOK(spline->intersectWith(acadLine, intType, 
							  *projPlane, points));
	    }

	    delete acadLine;
	}

	delete spline;
	break;

    default:
       return Acad::eInvalidInput;
    }

    return es;
}


static Acad::ErrorStatus intPline(const AcDbPoly*         poly, 
				  AcDb3dPolyline*         pline,
				  AcDb::Intersect         intType,
				  const AcGePlane*        projPlane,
				  AcGePoint3dArray&          points)
{
    Acad::ErrorStatus es = Acad::eOk;

    AcGePoint3dArray vertexArray;
    if ((es = poly->getVertices(vertexArray)) != Acad::eOk) {
        return es;
    }

    AcDbSpline*      spline = NULL;
    AcDbLine*	     acadLine = NULL;

    AcGePoint3dArray    pts;
    int              numPoints, i;

    AcGeLineSeg3d    lnsg;
    AcDb::Intersect  type = kOnBothOperands;

    switch (pline->polyType()) {

    case k3dSimplePoly:

	if ((es = rx_scanPline(pline, pts)) != Acad::eOk) {
            return es;
        }
	numPoints = pts.length();

	// Intersect with each line segment of the polyline. 
	// Depending on the intType, extend the last segment.

	for (i = 0; i < numPoints - 1; i++) {

	    if (i == numPoints - 2)
		type = intType;

	    lnsg.set(pts[i], pts[i + 1]);
	    if ((es = intLine(poly, lnsg, type, projPlane, points)) 
                   != Acad::eOk) {
                return es;
            }
	}

    case k3dQuadSplinePoly:
    case k3dCubicSplinePoly:

	if ((es = pline->getSpline(spline)) != Acad::eOk) {
            delete spline;
            return es;
        }

	if (intType == kExtendArg || intType == kExtendBoth) {
	    intType = kExtendThis;
	}

	for (i = 0; i < vertexArray.length() - 1; i++) {

	    acadLine = new AcDbLine();
	    acadLine->setStartPoint(vertexArray[i]);
	    acadLine->setEndPoint(vertexArray[i + 1]);
	    acadLine->setNormal(poly->normal());

	    if (projPlane == NULL) {
		AOK(spline->intersectWith(acadLine, intType, points));
	    } else {
		AOK(spline->intersectWith(acadLine, intType, 
							  *projPlane, points));
	    }

	    delete acadLine;
	}

	delete spline;
	break;

    default:
       return Acad::eInvalidInput;
    }

    return es;
}





ACRX_CONS_DEFINE_MEMBERS(AcDbPolyEditorReactor, AcEditorReactor, 0);;

static int numStarted = 0;

void
AcDbPolyEditorReactor::commandWillStart(const char * cmdStr)
{
#if 0
    if (!(!strcmp(cmdStr, "UNDO") || 
          !strcmp(cmdStr, "U")    || 
          !strcmp(cmdStr, "REDO"))) {
       AcTransaction* tr = actrTransactionManager->startTransaction();
    }
#endif
}

void
AcDbPolyEditorReactor::commandEnded(const char * cmdStr)
{
#if 0
    if (!(!strcmp(cmdStr, "UNDO") || 
          !strcmp(cmdStr, "U")    || 
          !strcmp(cmdStr, "REDO"))) {

        if (actrTransactionManager->numActiveTransactions() > 0)
            actrTransactionManager->endTransaction();
    }

    // Print the user typed name.
    // Get both the global name and local name and print that out too.
    //
    ads_printf("\nUser entered : %s", cmdStr);

    AcEdCommand* pCmd = NULL;
    if ((pCmd = acedRegCmds->lookupCmd(cmdStr, Adesk::kTrue, 
                                               Adesk::kTrue)) != NULL) {
         ads_printf("\nGlobal name : %s", pCmd->globalName());
         ads_printf("\nLocal  name : %s", pCmd->localName());
    } else if ((pCmd = acedRegCmds->lookupCmd(cmdStr, Adesk::kFalse, 
                                                      Adesk::kTrue)) != NULL) {
         ads_printf("\nGlobal name : %s", pCmd->globalName());
         ads_printf("\nLocal  name : %s", pCmd->localName());
    } else {
         ads_printf("\nThis is not a rx-registered command");
    }
#endif
}

void
AcDbPolyEditorReactor::commandCancelled(const char * cmdStr)
{
    if (!(!strcmp(cmdStr, "UNDO") || 
          !strcmp(cmdStr, "U")    || 
          !strcmp(cmdStr, "REDO"))) {

        if (actrTransactionManager->numActiveTransactions() > 0)
            actrTransactionManager->endTransaction();
    }
}

void
AcDbPolyEditorReactor::commandFailed(const char * cmdStr)
{
    if (!(!strcmp(cmdStr, "UNDO") || 
          !strcmp(cmdStr, "U")    || 
          !strcmp(cmdStr, "REDO"))) {

        if (actrTransactionManager->numActiveTransactions() > 0)
            actrTransactionManager->endTransaction();
    }
}

void
AcDbPolyEditorReactor::beginClose(AcDbDatabase* pDwg)
{
extern AcDbPolyDBReactor            *pDBReactor;
    if (pDwg == acdbCurDwg()) {
        pDwg->removeReactor(pDBReactor);
        delete pDBReactor;
        pDBReactor = NULL;
    }
}

void 
AcDbPolyEditorReactor::beginSave(AcDbDatabase* pDwg, const char* pIntendedName)
{
}


void 
AcDbPolyEditorReactor::saveComplete(AcDbDatabase* pDwg, const char* pActualName)
{
}


ACRX_CONS_DEFINE_MEMBERS(AcDbPolyDBReactor, AcDbDatabaseReactor, 0);;

void
AcDbPolyDBReactor::objectAppended(const AcDbDatabase* pDb,
                                const AcDbObject*   pObj)
{
    if (pObj->isKindOf(AcDbBlockReference::desc())) {
        ads_printf("\nWell, it actually is a block reference!!!");
    } else {
        AcRxClass* classObj = pObj->isA();
        //ads_printf("\nClass type is : %s", classObj->name());
    }
}


void
AcDbPolyDBReactor::objectUnAppended(const AcDbDatabase* pDb,
                                const AcDbObject*   pObj)
{
    ads_printf("\nDatabase UnAppended.");
}


void
AcDbPolyDBReactor::objectReAppended(const AcDbDatabase* pDb,
                                const AcDbObject*   pObj)
{
    ads_printf("\nDatabase ReAppended.");
}


ACRX_CONS_DEFINE_MEMBERS(AcDbPolyObjReactor, AcDbObjectReactor, 0);;

void	      
AcDbPolyObjReactor::erased(const AcDbObject* pObj,
		           Adesk::Boolean erasing)
{
    AcRxClass* classObj = pObj->isA();
#if 0
    if (erasing) 
        ads_printf("\nAn object of type %s is getting erased", 
					classObj->name());
    else
        ads_printf("\nAn object of type %s is getting unerased", 
					classObj->name());
#endif
}

void	      
AcDbPolyObjReactor::unappended(const AcDbObject* pObj)
{
    ads_printf("\nunappended.");
}

void	      
AcDbPolyObjReactor::reappended(const AcDbObject* pObj)
{
    ads_printf("\nreappended.");
}
